@@ -2,6 +2,7 @@
 #include "CSE.h"
 #include "CodeGen_Internal.h"
 #include "ConciseCasts.h"
+#include "ConstantBounds.h"
 #include "IRMatch.h"
 #include "IRMutator.h"
 #include "Simplify.h"
@@ -13,6 +14,8 @@ using namespace Halide::ConciseCasts;
 
 namespace {
 
+// This routine provides a guard on the return type of intrisics such that only
+// these types will ever be considered in the visiting that happens here.
 bool find_intrinsics_for_type(const Type &t) {
     // Currently, we only try to find and replace intrinsics for vector types that aren't bools.
     return t.is_vector() && t.bits() >= 8;
@@ -28,17 +31,19 @@ Expr narrow(Expr a) {
     return Cast::make(result_type, std::move(a));
 }
 
-Expr lossless_narrow(const Expr &x) {
-    return lossless_cast(x.type().narrow(), x);
-}
-
-// Remove a widening cast even if it changes the sign of the result.
-Expr strip_widening_cast(const Expr &x) {
-    Expr narrow = lossless_narrow(x);
-    if (narrow.defined()) {
-        return narrow;
-    }
-    return lossless_cast(x.type().narrow().with_code(halide_type_uint), x);
+// Check a type to make sure it can be narrowed. find_intrinsics_for_type
+// attempts to prevent this code from narrowing in cases that do not work, but
+// it is incomplete for two reasons:
+//
+//     - Arguments can be narrowed and that guard is only on return type, which
+//     are different for widening operations.
+//
+//     - find_intrinsics_for_type does not cull out float16, and it probably
+//     should not as while it's ok to skip matching bool things, float16 things
+//     are useful.
+bool can_narrow(const Type &t) {
+    return (t.is_float() && t.bits() >= 32) ||
+           t.bits() >= 8;
 }
 
 Expr saturating_narrow(const Expr &a) {
@@ -56,34 +61,6 @@ bool no_overflow(Type t) {
     return t.is_float() || no_overflow_int(t);
 }
 
-// If there's a widening add or subtract in the first e.type().bits() / 2 - 1
-// levels down a tree of adds or subtracts, we know there's enough headroom for
-// another add without overflow. For example, it is safe to add to
-// (widening_add(x, y) - z) without overflow.
-bool is_safe_for_add(const Expr &e, int max_depth) {
-    if (max_depth-- <= 0) {
-        return false;
-    }
-    if (const Add *add = e.as<Add>()) {
-        return is_safe_for_add(add->a, max_depth) || is_safe_for_add(add->b, max_depth);
-    } else if (const Sub *sub = e.as<Sub>()) {
-        return is_safe_for_add(sub->a, max_depth) || is_safe_for_add(sub->b, max_depth);
-    } else if (const Cast *cast = e.as<Cast>()) {
-        if (cast->type.bits() > cast->value.type().bits()) {
-            return true;
-        } else if (cast->type.bits() == cast->value.type().bits()) {
-            return is_safe_for_add(cast->value, max_depth);
-        }
-    } else if (Call::as_intrinsic(e, {Call::widening_add, Call::widening_sub, Call::widen_right_add, Call::widen_right_sub})) {
-        return true;
-    }
-    return false;
-}
-
-bool is_safe_for_add(const Expr &e) {
-    return is_safe_for_add(e, e.type().bits() / 2 - 1);
-}
-
 // We want to find and remove an add of 'round' from e. This is not
 // the same thing as just subtracting round, we specifically want
 // to remove an addition of exactly round.
@@ -109,93 +86,6 @@ Expr find_and_subtract(const Expr &e, const Expr &round) {
     return Expr();
 }
 
-Expr to_rounding_shift(const Call *c) {
-    if (c->is_intrinsic(Call::shift_left) || c->is_intrinsic(Call::shift_right)) {
-        internal_assert(c->args.size() == 2);
-        Expr a = c->args[0];
-        Expr b = c->args[1];
-
-        // Helper to make the appropriate shift.
-        auto rounding_shift = [&](const Expr &a, const Expr &b) {
-            if (c->is_intrinsic(Call::shift_right)) {
-                return rounding_shift_right(a, b);
-            } else {
-                return rounding_shift_left(a, b);
-            }
-        };
-
-        // The rounding offset for the shift we have.
-        Type round_type = a.type().with_lanes(1);
-        if (Call::as_intrinsic(a, {Call::widening_add})) {
-            round_type = round_type.narrow();
-        }
-        Expr round;
-        if (c->is_intrinsic(Call::shift_right)) {
-            round = (make_one(round_type) << max(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2;
-        } else {
-            round = (make_one(round_type) >> min(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2;
-        }
-        // Input expressions are simplified before running find_intrinsics, but b
-        // has been lifted here so we need to lower_intrinsics before simplifying
-        // and re-lifting. Should we move this code into the FindIntrinsics class
-        // to make it easier to lift round?
-        round = lower_intrinsics(round);
-        round = simplify(round);
-        round = find_intrinsics(round);
-
-        // We can always handle widening adds.
-        if (const Call *add = Call::as_intrinsic(a, {Call::widening_add})) {
-            if (can_prove(lower_intrinsics(add->args[0] == round))) {
-                return rounding_shift(cast(add->type, add->args[1]), b);
-            } else if (can_prove(lower_intrinsics(add->args[1] == round))) {
-                return rounding_shift(cast(add->type, add->args[0]), b);
-            }
-        }
-
-        if (const Call *add = Call::as_intrinsic(a, {Call::widen_right_add})) {
-            if (can_prove(lower_intrinsics(add->args[1] == round))) {
-                return rounding_shift(cast(add->type, add->args[0]), b);
-            }
-        }
-
-        // Also need to handle the annoying case of a reinterpret cast wrapping a widen_right_add
-        // TODO: this pattern makes me want to change the semantics of this op.
-        if (const Cast *cast = a.as<Cast>()) {
-            if (cast->is_reinterpret()) {
-                if (const Call *add = Call::as_intrinsic(cast->value, {Call::widen_right_add})) {
-                    if (can_prove(lower_intrinsics(add->args[1] == round))) {
-                        // We expect the first operand to be a reinterpet cast.
-                        if (const Cast *cast_a = add->args[0].as<Cast>()) {
-                            if (cast_a->is_reinterpret()) {
-                                return rounding_shift(cast_a->value, b);
-                            }
-                        }
-                    }
-                }
-            }
-        }
-
-        // If it wasn't a widening or saturating add, we might still
-        // be able to safely accept the rounding.
-        Expr a_less_round = find_and_subtract(a, round);
-        if (a_less_round.defined()) {
-            // We found and removed the rounding. However, we may have just changed
-            // behavior due to overflow. This is still safe if the type is not
-            // overflowing, or we can find a widening add or subtract in the tree
-            // of adds/subtracts. This is a common pattern, e.g.
-            // rounding_halving_add(a, b) = shift_round(widening_add(a, b) + 1, 1).
-            // TODO: This could be done with bounds inference instead of this hack
-            // if it supported intrinsics like widening_add and tracked bounds for
-            // types other than int32.
-            if (no_overflow(a.type()) || is_safe_for_add(a_less_round)) {
-                return rounding_shift(simplify(a_less_round), b);
-            }
-        }
-    }
-
-    return Expr();
-}
-
 class FindIntrinsics : public IRMutator {
 protected:
     using IRMutator::visit;
@@ -207,6 +97,119 @@ protected:
     IRMatcher::WildConst<0> c0;
     IRMatcher::WildConst<1> c1;
 
+    std::map<Expr, ConstantInterval, ExprCompare> bounds_cache;
+    Scope<ConstantInterval> let_var_bounds;
+
+    Expr lossless_cast(Type t, const Expr &e) {
+        return Halide::Internal::lossless_cast(t, e, &bounds_cache);
+    }
+
+    ConstantInterval constant_integer_bounds(const Expr &e) {
+        // TODO: Use the scope - add let visitors
+        return Halide::Internal::constant_integer_bounds(e, let_var_bounds, &bounds_cache);
+    }
+
+    Expr lossless_narrow(const Expr &x) {
+        return can_narrow(x.type()) ? lossless_cast(x.type().narrow(), x) : Expr();
+    }
+
+    // Remove a widening cast even if it changes the sign of the result.
+    Expr strip_widening_cast(const Expr &x) {
+        if (can_narrow(x.type())) {
+            Expr narrow = lossless_narrow(x);
+            if (narrow.defined()) {
+                return narrow;
+            }
+            return lossless_cast(x.type().narrow().with_code(halide_type_uint), x);
+        } else {
+            return Expr();
+        }
+    }
+
+    Expr to_rounding_shift(const Call *c) {
+        if (c->is_intrinsic(Call::shift_left) || c->is_intrinsic(Call::shift_right)) {
+            internal_assert(c->args.size() == 2);
+            Expr a = c->args[0];
+            Expr b = c->args[1];
+
+            // Helper to make the appropriate shift.
+            auto rounding_shift = [&](const Expr &a, const Expr &b) {
+                if (c->is_intrinsic(Call::shift_right)) {
+                    return rounding_shift_right(a, b);
+                } else {
+                    return rounding_shift_left(a, b);
+                }
+            };
+
+            // The rounding offset for the shift we have.
+            Type round_type = a.type().with_lanes(1);
+            if (Call::as_intrinsic(a, {Call::widening_add})) {
+                round_type = round_type.narrow();
+            }
+            Expr round;
+            if (c->is_intrinsic(Call::shift_right)) {
+                round = (make_one(round_type) << max(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2;
+            } else {
+                round = (make_one(round_type) >> min(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2;
+            }
+            // Input expressions are simplified before running find_intrinsics, but b
+            // has been lifted here so we need to lower_intrinsics before simplifying
+            // and re-lifting. Should we move this code into the FindIntrinsics class
+            // to make it easier to lift round?
+            round = lower_intrinsics(round);
+            round = simplify(round);
+            round = find_intrinsics(round);
+
+            // We can always handle widening adds.
+            if (const Call *add = Call::as_intrinsic(a, {Call::widening_add})) {
+                if (can_prove(lower_intrinsics(add->args[0] == round))) {
+                    return rounding_shift(cast(add->type, add->args[1]), b);
+                } else if (can_prove(lower_intrinsics(add->args[1] == round))) {
+                    return rounding_shift(cast(add->type, add->args[0]), b);
+                }
+            }
+
+            if (const Call *add = Call::as_intrinsic(a, {Call::widen_right_add})) {
+                if (can_prove(lower_intrinsics(add->args[1] == round))) {
+                    return rounding_shift(cast(add->type, add->args[0]), b);
+                }
+            }
+
+            // Also need to handle the annoying case of a reinterpret cast wrapping a widen_right_add
+            if (const Cast *cast = a.as<Cast>()) {
+                if (cast->is_reinterpret()) {
+                    if (const Call *add = Call::as_intrinsic(cast->value, {Call::widen_right_add})) {
+                        if (can_prove(lower_intrinsics(add->args[1] == round))) {
+                            // We expect the first operand to be a reinterpet cast.
+                            if (const Cast *cast_a = add->args[0].as<Cast>()) {
+                                if (cast_a->is_reinterpret()) {
+                                    return rounding_shift(cast_a->value, b);
+                                }
+                            }
+                        }
+                    }
+                }
+            }
+
+            // If it wasn't a widening or saturating add, we might still
+            // be able to safely accept the rounding.
+            Expr a_less_round = find_and_subtract(a, round);
+            if (a_less_round.defined()) {
+                // We found and removed the rounding. Verify it didn't change
+                // overflow behavior.
+                if (no_overflow(a.type()) ||
+                    a.type().can_represent(constant_integer_bounds(a_less_round) +
+                                           constant_integer_bounds(round))) {
+                    // If we can add the rounding term back on without causing
+                    // overflow, then it must not have overflowed originally.
+                    return rounding_shift(simplify(a_less_round), b);
+                }
+            }
+        }
+
+        return Expr();
+    }
+
     Expr visit(const Add *op) override {
         if (!find_intrinsics_for_type(op->type)) {
             return IRMutator::visit(op);
@@ -217,16 +220,18 @@ protected:
 
         // Try widening both from the same signedness as the result, and from uint.
         for (halide_type_code_t code : {op->type.code(), halide_type_uint}) {
-            Type narrow = op->type.narrow().with_code(code);
-            Expr narrow_a = lossless_cast(narrow, a);
-            Expr narrow_b = lossless_cast(narrow, b);
+            if (can_narrow(op->type)) {
+                Type narrow = op->type.narrow().with_code(code);
+                Expr narrow_a = lossless_cast(narrow, a);
+                Expr narrow_b = lossless_cast(narrow, b);
 
-            if (narrow_a.defined() && narrow_b.defined()) {
-                Expr result = widening_add(narrow_a, narrow_b);
-                if (result.type() != op->type) {
-                    result = Cast::make(op->type, result);
+                if (narrow_a.defined() && narrow_b.defined()) {
+                    Expr result = widening_add(narrow_a, narrow_b);
+                    if (result.type() != op->type) {
+                        result = Cast::make(op->type, result);
+                    }
+                    return mutate(result);
                 }
-                return mutate(result);
             }
         }
 
@@ -235,41 +240,43 @@ protected:
             // Yes we do an duplicate code, but we want to check the op->type.code() first,
             // and the opposite as well.
             for (halide_type_code_t code : {op->type.code(), halide_type_uint, halide_type_int}) {
-                Type narrow = op->type.narrow().with_code(code);
-                // Pulling casts out of VectorReduce nodes breaks too much codegen, skip for now.
-                Expr narrow_a = (a.node_type() == IRNodeType::VectorReduce) ? Expr() : lossless_cast(narrow, a);
-                Expr narrow_b = (b.node_type() == IRNodeType::VectorReduce) ? Expr() : lossless_cast(narrow, b);
+                if (can_narrow(op->type)) {
+                    Type narrow = op->type.narrow().with_code(code);
+                    // Pulling casts out of VectorReduce nodes breaks too much codegen, skip for now.
+                    Expr narrow_a = (a.node_type() == IRNodeType::VectorReduce) ? Expr() : lossless_cast(narrow, a);
+                    Expr narrow_b = (b.node_type() == IRNodeType::VectorReduce) ? Expr() : lossless_cast(narrow, b);
 
-                // This case should have been handled by the above check for widening_add.
-                internal_assert(!(narrow_a.defined() && narrow_b.defined()))
-                    << "find_intrinsics failed to find a widening_add: " << a << " + " << b << "\n";
+                    // This case should have been handled by the above check for widening_add.
+                    internal_assert(!(narrow_a.defined() && narrow_b.defined()))
+                        << "find_intrinsics failed to find a widening_add: " << a << " + " << b << "\n";
 
-                if (narrow_a.defined()) {
-                    Expr result;
-                    if (b.type().code() != narrow_a.type().code()) {
-                        // Need to do a safe reinterpret.
-                        Type t = b.type().with_code(code);
-                        result = widen_right_add(cast(t, b), narrow_a);
-                        internal_assert(result.type() != op->type);
-                        result = cast(op->type, result);
-                    } else {
-                        result = widen_right_add(b, narrow_a);
+                    if (narrow_a.defined()) {
+                        Expr result;
+                        if (b.type().code() != narrow_a.type().code()) {
+                            // Need to do a safe reinterpret.
+                            Type t = b.type().with_code(code);
+                            result = widen_right_add(cast(t, b), narrow_a);
+                            internal_assert(result.type() != op->type);
+                            result = cast(op->type, result);
+                        } else {
+                            result = widen_right_add(b, narrow_a);
+                        }
+                        internal_assert(result.type() == op->type);
+                        return mutate(result);
+                    } else if (narrow_b.defined()) {
+                        Expr result;
+                        if (a.type().code() != narrow_b.type().code()) {
+                            // Need to do a safe reinterpret.
+                            Type t = a.type().with_code(code);
+                            result = widen_right_add(cast(t, a), narrow_b);
+                            internal_assert(result.type() != op->type);
+                            result = cast(op->type, result);
+                        } else {
+                            result = widen_right_add(a, narrow_b);
+                        }
+                        internal_assert(result.type() == op->type);
+                        return mutate(result);
                     }
-                    internal_assert(result.type() == op->type);
-                    return mutate(result);
-                } else if (narrow_b.defined()) {
-                    Expr result;
-                    if (a.type().code() != narrow_b.type().code()) {
-                        // Need to do a safe reinterpret.
-                        Type t = a.type().with_code(code);
-                        result = widen_right_add(cast(t, a), narrow_b);
-                        internal_assert(result.type() != op->type);
-                        result = cast(op->type, result);
-                    } else {
-                        result = widen_right_add(a, narrow_b);
-                    }
-                    internal_assert(result.type() == op->type);
-                    return mutate(result);
                 }
             }
         }
@@ -294,22 +301,24 @@ protected:
 
         // Try widening both from the same type as the result, and from uint.
         for (halide_type_code_t code : {op->type.code(), halide_type_uint}) {
-            Type narrow = op->type.narrow().with_code(code);
-            Expr narrow_a = lossless_cast(narrow, a);
-            Expr narrow_b = lossless_cast(narrow, b);
+            if (can_narrow(op->type)) {
+                Type narrow = op->type.narrow().with_code(code);
+                Expr narrow_a = lossless_cast(narrow, a);
+                Expr narrow_b = lossless_cast(narrow, b);
 
-            if (narrow_a.defined() && narrow_b.defined()) {
-                Expr negative_narrow_b = lossless_negate(narrow_b);
-                Expr result;
-                if (negative_narrow_b.defined()) {
-                    result = widening_add(narrow_a, negative_narrow_b);
-                } else {
-                    result = widening_sub(narrow_a, narrow_b);
+                if (narrow_a.defined() && narrow_b.defined()) {
+                    Expr negative_narrow_b = lossless_negate(narrow_b);
+                    Expr result;
+                    if (negative_narrow_b.defined()) {
+                        result = widening_add(narrow_a, negative_narrow_b);
+                    } else {
+                        result = widening_sub(narrow_a, narrow_b);
+                    }
+                    if (result.type() != op->type) {
+                        result = Cast::make(op->type, result);
+                    }
+                    return mutate(result);
                 }
-                if (result.type() != op->type) {
-                    result = Cast::make(op->type, result);
-                }
-                return mutate(result);
             }
         }
 
@@ -324,22 +333,24 @@ protected:
             // Yes we do an duplicate code, but we want to check the op->type.code() first,
             // and the opposite as well.
             for (halide_type_code_t code : {op->type.code(), halide_type_uint, halide_type_int}) {
-                Type narrow = op->type.narrow().with_code(code);
-                Expr narrow_b = lossless_cast(narrow, b);
+                if (can_narrow(op->type)) {
+                    Type narrow = op->type.narrow().with_code(code);
+                    Expr narrow_b = lossless_cast(narrow, b);
 
-                if (narrow_b.defined()) {
-                    Expr result;
-                    if (a.type().code() != narrow_b.type().code()) {
-                        // Need to do a safe reinterpret.
-                        Type t = a.type().with_code(code);
-                        result = widen_right_sub(cast(t, a), narrow_b);
-                        internal_assert(result.type() != op->type);
-                        result = cast(op->type, result);
-                    } else {
-                        result = widen_right_sub(a, narrow_b);
+                    if (narrow_b.defined()) {
+                        Expr result;
+                        if (a.type().code() != narrow_b.type().code()) {
+                            // Need to do a safe reinterpret.
+                            Type t = a.type().with_code(code);
+                            result = widen_right_sub(cast(t, a), narrow_b);
+                            internal_assert(result.type() != op->type);
+                            result = cast(op->type, result);
+                        } else {
+                            result = widen_right_sub(a, narrow_b);
+                        }
+                        internal_assert(result.type() == op->type);
+                        return mutate(result);
                     }
-                    internal_assert(result.type() == op->type);
-                    return mutate(result);
                 }
             }
         }
@@ -401,40 +412,42 @@ protected:
             // Yes we do an duplicate code, but we want to check the op->type.code() first,
             // and the opposite as well.
             for (halide_type_code_t code : {op->type.code(), halide_type_uint, halide_type_int}) {
-                Type narrow = op->type.narrow().with_code(code);
-                Expr narrow_a = lossless_cast(narrow, a);
-                Expr narrow_b = lossless_cast(narrow, b);
+                if (can_narrow(op->type)) {
+                    Type narrow = op->type.narrow().with_code(code);
+                    Expr narrow_a = lossless_cast(narrow, a);
+                    Expr narrow_b = lossless_cast(narrow, b);
 
-                // This case should have been handled by the above check for widening_mul.
-                internal_assert(!(narrow_a.defined() && narrow_b.defined()))
-                    << "find_intrinsics failed to find a widening_mul: " << a << " + " << b << "\n";
+                    // This case should have been handled by the above check for widening_mul.
+                    internal_assert(!(narrow_a.defined() && narrow_b.defined()))
+                        << "find_intrinsics failed to find a widening_mul: " << a << " + " << b << "\n";
 
-                if (narrow_a.defined()) {
-                    Expr result;
-                    if (b.type().code() != narrow_a.type().code()) {
-                        // Need to do a safe reinterpret.
-                        Type t = b.type().with_code(code);
-                        result = widen_right_mul(cast(t, b), narrow_a);
-                        internal_assert(result.type() != op->type);
-                        result = cast(op->type, result);
-                    } else {
-                        result = widen_right_mul(b, narrow_a);
+                    if (narrow_a.defined()) {
+                        Expr result;
+                        if (b.type().code() != narrow_a.type().code()) {
+                            // Need to do a safe reinterpret.
+                            Type t = b.type().with_code(code);
+                            result = widen_right_mul(cast(t, b), narrow_a);
+                            internal_assert(result.type() != op->type);
+                            result = cast(op->type, result);
+                        } else {
+                            result = widen_right_mul(b, narrow_a);
+                        }
+                        internal_assert(result.type() == op->type);
+                        return mutate(result);
+                    } else if (narrow_b.defined()) {
+                        Expr result;
+                        if (a.type().code() != narrow_b.type().code()) {
+                            // Need to do a safe reinterpret.
+                            Type t = a.type().with_code(code);
+                            result = widen_right_mul(cast(t, a), narrow_b);
+                            internal_assert(result.type() != op->type);
+                            result = cast(op->type, result);
+                        } else {
+                            result = widen_right_mul(a, narrow_b);
+                        }
+                        internal_assert(result.type() == op->type);
+                        return mutate(result);
                     }
-                    internal_assert(result.type() == op->type);
-                    return mutate(result);
-                } else if (narrow_b.defined()) {
-                    Expr result;
-                    if (a.type().code() != narrow_b.type().code()) {
-                        // Need to do a safe reinterpret.
-                        Type t = a.type().with_code(code);
-                        result = widen_right_mul(cast(t, a), narrow_b);
-                        internal_assert(result.type() != op->type);
-                        result = cast(op->type, result);
-                    } else {
-                        result = widen_right_mul(a, narrow_b);
-                    }
-                    internal_assert(result.type() == op->type);
-                    return mutate(result);
                 }
             }
         }
@@ -517,6 +530,11 @@ protected:
             }
         }
 
+        // Do we need to worry about this cast overflowing?
+        ConstantInterval value_bounds = constant_integer_bounds(value);
+        bool no_overflow = (op->type.can_represent(op->value.type()) ||
+                            op->type.can_represent(value_bounds));
+
         if (op->type.is_int() || op->type.is_uint()) {
             Expr lower = cast(value.type(), op->type.min());
             Expr upper = cast(value.type(), op->type.max());
@@ -534,7 +552,6 @@ protected:
             auto is_x_same_uint = op->type.is_uint() && is_uint(x, bits);
             auto is_x_same_int_or_uint = is_x_same_int || is_x_same_uint;
             auto x_y_same_sign = (is_int(x) && is_int(y)) || (is_uint(x) && is_uint(y));
-            auto is_y_narrow_uint = op->type.is_uint() && is_uint(y, bits / 2);
             if (
                 // Saturating patterns
                 rewrite(max(min(widening_add(x, y), upper), lower),
@@ -636,32 +653,16 @@ protected:
                         rounding_mul_shift_right(x, y, cast(unsigned_type, c0)),
                         is_x_same_int && x_y_same_sign && c0 >= bits - 1) ||
 
-                rewrite(shift_right(widening_mul(x, y), c0),
-                        mul_shift_right(x, y, cast(unsigned_type, c0)),
-                        is_x_same_int_or_uint && x_y_same_sign && c0 >= bits) ||
+                // We can also match whenever the cast can't overflow, so
+                // questions of saturation are irrelevant.
+                (no_overflow &&
+                 (rewrite(shift_right(widening_mul(x, y), c0),
+                          mul_shift_right(x, y, cast(unsigned_type, c0)),
+                          is_x_same_int_or_uint && x_y_same_sign && c0 >= 0) ||
 
-                rewrite(rounding_shift_right(widening_mul(x, y), c0),
-                        rounding_mul_shift_right(x, y, cast(unsigned_type, c0)),
-                        is_x_same_int_or_uint && x_y_same_sign && c0 >= bits) ||
-
-                // We can also match on smaller shifts if one of the args is
-                // narrow. We don't do this for signed (yet), because the
-                // saturation issue is tricky.
-                rewrite(shift_right(widening_mul(x, cast(op->type, y)), c0),
-                        mul_shift_right(x, cast(op->type, y), cast(unsigned_type, c0)),
-                        is_x_same_int_or_uint && is_y_narrow_uint && c0 >= bits / 2) ||
-
-                rewrite(rounding_shift_right(widening_mul(x, cast(op->type, y)), c0),
-                        rounding_mul_shift_right(x, cast(op->type, y), cast(unsigned_type, c0)),
-                        is_x_same_int_or_uint && is_y_narrow_uint && c0 >= bits / 2) ||
-
-                rewrite(shift_right(widening_mul(cast(op->type, y), x), c0),
-                        mul_shift_right(cast(op->type, y), x, cast(unsigned_type, c0)),
-                        is_x_same_int_or_uint && is_y_narrow_uint && c0 >= bits / 2) ||
-
-                rewrite(rounding_shift_right(widening_mul(cast(op->type, y), x), c0),
-                        rounding_mul_shift_right(cast(op->type, y), x, cast(unsigned_type, c0)),
-                        is_x_same_int_or_uint && is_y_narrow_uint && c0 >= bits / 2) ||
+                  rewrite(rounding_shift_right(widening_mul(x, y), c0),
+                          rounding_mul_shift_right(x, y, cast(unsigned_type, c0)),
+                          is_x_same_int_or_uint && x_y_same_sign && c0 >= 0))) ||
 
                 // Halving subtract patterns
                 rewrite(shift_right(cast(op_type_wide, widening_sub(x, y)), 1),
@@ -853,33 +854,40 @@ protected:
         } else if (op->is_intrinsic(Call::widening_add) && (op->type.bits() >= 16)) {
             internal_assert(op->args.size() == 2);
             for (halide_type_code_t t : {op->type.code(), halide_type_uint}) {
-                Type narrow_t = op->type.narrow().narrow().with_code(t);
-                Expr narrow_a = lossless_cast(narrow_t, op->args[0]);
-                Expr narrow_b = lossless_cast(narrow_t, op->args[1]);
-                if (narrow_a.defined() && narrow_b.defined()) {
-                    return mutate(Cast::make(op->type, widening_add(narrow_a, narrow_b)));
+                if (can_narrow(op->type)) {
+                    Type narrow_t = op->type.narrow().narrow().with_code(t);
+                    Expr narrow_a = lossless_cast(narrow_t, op->args[0]);
+                    Expr narrow_b = lossless_cast(narrow_t, op->args[1]);
+                    if (narrow_a.defined() && narrow_b.defined()) {
+                        return mutate(Cast::make(op->type, widening_add(narrow_a, narrow_b)));
+                    }
                 }
             }
         } else if (op->is_intrinsic(Call::widening_sub) && (op->type.bits() >= 16)) {
             internal_assert(op->args.size() == 2);
             for (halide_type_code_t t : {op->type.code(), halide_type_uint}) {
-                Type narrow_t = op->type.narrow().narrow().with_code(t);
-                Expr narrow_a = lossless_cast(narrow_t, op->args[0]);
-                Expr narrow_b = lossless_cast(narrow_t, op->args[1]);
-                if (narrow_a.defined() && narrow_b.defined()) {
-                    return mutate(Cast::make(op->type, widening_sub(narrow_a, narrow_b)));
+                if (can_narrow(op->type)) {
+                    Type narrow_t = op->type.narrow().narrow().with_code(t);
+                    Expr narrow_a = lossless_cast(narrow_t, op->args[0]);
+                    Expr narrow_b = lossless_cast(narrow_t, op->args[1]);
+                    if (narrow_a.defined() && narrow_b.defined()) {
+                        return mutate(Cast::make(op->type, widening_sub(narrow_a, narrow_b)));
+                    }
                 }
             }
         }
         // TODO: do we want versions of widen_right_add here?
 
-        if (op->is_intrinsic(Call::shift_right) || op->is_intrinsic(Call::shift_left)) {
+        if (op->is_intrinsic(Call::shift_right) ||
+            op->is_intrinsic(Call::shift_left)) {
             // Try to turn this into a widening shift.
             internal_assert(op->args.size() == 2);
             Expr a_narrow = lossless_narrow(op->args[0]);
             Expr b_narrow = lossless_narrow(op->args[1]);
             if (a_narrow.defined() && b_narrow.defined()) {
-                Expr result = op->is_intrinsic(Call::shift_left) ? widening_shift_left(a_narrow, b_narrow) : widening_shift_right(a_narrow, b_narrow);
+                Expr result = op->is_intrinsic(Call::shift_left) ?
+                                  widening_shift_left(a_narrow, b_narrow) :
+                                  widening_shift_right(a_narrow, b_narrow);
                 if (result.type() != op->type) {
                     result = Cast::make(op->type, result);
                 }
@@ -893,16 +901,20 @@ protected:
             }
         }
 
-        if (op->is_intrinsic(Call::rounding_shift_left) || op->is_intrinsic(Call::rounding_shift_right)) {
+        if (op->is_intrinsic(Call::rounding_shift_left) ||
+            op->is_intrinsic(Call::rounding_shift_right)) {
             // Try to turn this into a widening shift.
             internal_assert(op->args.size() == 2);
             Expr a_narrow = lossless_narrow(op->args[0]);
             Expr b_narrow = lossless_narrow(op->args[1]);
             if (a_narrow.defined() && b_narrow.defined()) {
                 Expr result;
-                if (op->is_intrinsic(Call::rounding_shift_right) && can_prove(b_narrow > 0)) {
+                if (op->is_intrinsic(Call::rounding_shift_right) &&
+                    can_prove(b_narrow > 0 && b_narrow < a_narrow.type().bits())) {
                     result = rounding_shift_right(a_narrow, b_narrow);
-                } else if (op->is_intrinsic(Call::rounding_shift_left) && can_prove(b_narrow < 0)) {
+                } else if (op->is_intrinsic(Call::rounding_shift_left) &&
+                           b_narrow.type().is_int() &&
+                           can_prove(b_narrow < 0 && b_narrow > -a_narrow.type().bits())) {
                     result = rounding_shift_left(a_narrow, b_narrow);
                 } else {
                     return op;
@@ -1083,8 +1095,8 @@ class SubstituteInWideningLets : public IRMutator {
 
     Scope<Expr> replacements;
     Expr visit(const Variable *op) override {
-        if (replacements.contains(op->name)) {
-            return replacements.get(op->name);
+        if (const Expr *e = replacements.find(op->name)) {
+            return *e;
         } else {
             return op;
         }
@@ -1239,10 +1251,11 @@ Expr lower_widening_shift_right(const Expr &a, const Expr &b) {
 }
 
 Expr lower_rounding_shift_left(const Expr &a, const Expr &b) {
-    // Shift left, then add one to the result if bits were dropped
-    // (because b < 0) and the most significant dropped bit was a one.
+    // Shift left, then add one to the result if bits were dropped (because b < 0)
+    // and the most significant dropped bit was a one. We must take care not
+    // to introduce UB in the shifts, even if the result would be masked off.
     Expr b_negative = select(b < 0, make_one(a.type()), make_zero(a.type()));
-    return simplify((a << b) + (b_negative & (a << (b + 1))));
+    return simplify((a << b) + (b_negative & (a << saturating_add(b, make_one(b.type())))));
 }
 
 Expr lower_rounding_shift_right(const Expr &a, const Expr &b) {
@@ -1254,10 +1267,11 @@ Expr lower_rounding_shift_right(const Expr &a, const Expr &b) {
         Expr round = simplify(cast(a.type(), (1 << shift) - 1));
         return rounding_halving_add(a, round) >> shift;
     }
-    // Shift right, then add one to the result if bits were dropped
-    // (because b > 0) and the most significant dropped bit was a one.
+    // Shift right, then add one to the result if bits were dropped (because b > 0)
+    // and the most significant dropped bit was a one.  We must take care not to
+    // introduce UB in the shifts, even if the result would be masked off.
     Expr b_positive = select(b > 0, make_one(a.type()), make_zero(a.type()));
-    return simplify((a >> b) + (b_positive & (a >> (b - 1))));
+    return simplify((a >> b) + (b_positive & (a >> saturating_sub(b, make_one(b.type())))));
 }
 
 Expr lower_saturating_add(const Expr &a, const Expr &b) {
@@ -1450,27 +1464,45 @@ Expr lower_rounding_mul_shift_right(const Expr &a, const Expr &b, const Expr &q)
     // one of the operands and the denominator by a constant. We only do this
     // if it isn't already full precision. This avoids infinite loops despite
     // "lowering" this to another mul_shift_right operation.
-    if (can_prove(q < full_q)) {
-        Expr missing_q = full_q - q;
-        internal_assert(missing_q.type().bits() == b.type().bits());
-        Expr new_b = simplify(b << missing_q);
-        if (is_const(new_b) && can_prove(new_b >> missing_q == b)) {
-            return rounding_mul_shift_right(a, new_b, full_q);
+    ConstantInterval cq = constant_integer_bounds(q);
+    if (cq.is_single_point() && cq.max >= 0 && cq.max < full_q) {
+        int missing_q = full_q - (int)cq.max;
+
+        // Try to scale up the args by factors of two without overflowing
+        int a_shift = 0, b_shift = 0;
+        ConstantInterval ca = constant_integer_bounds(a);
+        while (true) {
+            ConstantInterval bigger = ca * 2;
+            if (a.type().can_represent(bigger) && a_shift + b_shift < missing_q) {
+                ca = bigger;
+                a_shift++;
+            } else {
+                break;
+            }
         }
-        Expr new_a = simplify(a << missing_q);
-        if (is_const(new_a) && can_prove(new_a >> missing_q == a)) {
-            return rounding_mul_shift_right(new_a, b, full_q);
+        ConstantInterval cb = constant_integer_bounds(b);
+        while (true) {
+            ConstantInterval bigger = cb * 2;
+            if (b.type().can_represent(bigger) && a_shift + b_shift < missing_q) {
+                cb = bigger;
+                b_shift++;
+            } else {
+                break;
+            }
+        }
+        if (a_shift + b_shift == missing_q) {
+            return rounding_mul_shift_right(simplify(a << a_shift), simplify(b << b_shift), full_q);
         }
     }
 
     // If all else fails, just widen, shift, and narrow.
-    Expr result = rounding_shift_right(widening_mul(a, b), q);
-    if (!can_prove(q >= a.type().bits())) {
-        result = saturating_narrow(result);
+    Expr wide_result = rounding_shift_right(widening_mul(a, b), q);
+    Expr narrowed = lossless_cast(a.type(), wide_result);
+    if (narrowed.defined()) {
+        return narrowed;
     } else {
-        result = narrow(result);
+        return saturating_narrow(wide_result);
     }
-    return result;
 }
 
 Expr lower_intrinsic(const Call *op) {
