@@ -4,6 +4,7 @@
 
 #include "Bounds.h"
 #include "CSE.h"
+#include "CanonicalizeGPUVars.h"
 #include "CodeGen_GPU_Dev.h"
 #include "CompilerLogger.h"
 #include "ExprUsesVar.h"
@@ -29,17 +30,14 @@ using std::vector;
 
 namespace {
 
-string thread_names[] = {"__thread_id_x", "__thread_id_y", "__thread_id_z", "__thread_id_w"};
-string block_names[] = {"__block_id_x", "__block_id_y", "__block_id_z", "__block_id_w"};
-
 class ExtractBlockSize : public IRVisitor {
-    Expr block_extent[4], block_count[4];
-    string block_var_name[4];
+    Expr block_extent[3], block_count[3];
+    string block_var_name[3];
 
     using IRVisitor::visit;
 
     void found_thread_for(int dim, const string &name, const Expr &extent) {
-        internal_assert(dim >= 0 && dim < 4);
+        internal_assert(dim >= 0 && dim < 3);
         if (!block_extent[dim].defined()) {
             block_extent[dim] = extent;
         } else {
@@ -48,17 +46,17 @@ class ExtractBlockSize : public IRVisitor {
     }
 
     void found_block_for(int dim, const string &name, Expr extent) {
-        internal_assert(dim >= 0 && dim < 4);
+        internal_assert(dim >= 0 && dim < 3);
         internal_assert(!block_count[dim].defined());
         block_count[dim] = std::move(extent);
         block_var_name[dim] = name;
     }
 
     void visit(const For *op) override {
-        for (int i = 0; i < 4; i++) {
-            if (ends_with(op->name, thread_names[i])) {
+        for (int i = 0; i < 3; i++) {
+            if (ends_with(op->name, gpu_thread_name(i))) {
                 found_thread_for(i, op->name, op->extent);
-            } else if (ends_with(op->name, block_names[i])) {
+            } else if (ends_with(op->name, gpu_block_name(i))) {
                 found_block_for(i, op->name, op->extent);
             }
         }
@@ -88,21 +86,21 @@ class ExtractBlockSize : public IRVisitor {
 
 public:
     int blocks_dimensions() const {
-        for (int i = 0; i < 4; i++) {
+        for (int i = 0; i < 3; i++) {
             if (!block_count[i].defined()) {
                 return i;
             }
         }
-        return 4;
+        return 3;
     }
 
     int threads_dimensions() const {
-        for (int i = 0; i < 4; i++) {
+        for (int i = 0; i < 3; i++) {
             if (!block_extent[i].defined()) {
                 return i;
             }
         }
-        return 4;
+        return 3;
     }
 
     Expr num_threads(int d) const {
@@ -114,12 +112,13 @@ public:
     }
 
     Expr block_var(int d) const {
+        // The name of the actual for loop
         return Variable::make(Int(32), block_var_name[d]);
     }
 
     Expr thread_var(int d) const {
         // Thread variables get canonical names
-        return Variable::make(Int(32), "." + thread_names[d]);
+        return Variable::make(Int(32), gpu_thread_name(d));
     }
 };
 
@@ -142,8 +141,8 @@ class NormalizeDimensionality : public IRMutator {
             return s;
         }
         while (max_depth < block_size.threads_dimensions()) {
-            string name = thread_names[max_depth];
-            s = For::make("." + name, 0, 1, ForType::GPUThread, Partition::Never, device_api, s);
+            s = For::make(gpu_thread_name(max_depth), 0, 1, ForType::GPUThread,
+                          Partition::Never, device_api, s);
             max_depth++;
         }
         return s;
@@ -166,7 +165,8 @@ class NormalizeDimensionality : public IRMutator {
     }
 
     Stmt visit(const For *op) override {
-        if (CodeGen_GPU_Dev::is_gpu_thread_var(op->name)) {
+        if (op->for_type == ForType::GPUThread ||
+            op->for_type == ForType::GPULane) {
             depth++;
             if (depth > max_depth) {
                 max_depth = depth;
@@ -191,10 +191,11 @@ class ReplaceForWithIf : public IRMutator {
     const ExtractBlockSize &block_size;
 
     Stmt visit(const For *op) override {
-        if (CodeGen_GPU_Dev::is_gpu_thread_var(op->name)) {
+        if (op->for_type == ForType::GPUThread ||
+            op->for_type == ForType::GPULane) {
             int dim;
-            for (dim = 0; dim < 4; dim++) {
-                if (ends_with(op->name, thread_names[dim])) {
+            for (dim = 0; dim < 3; dim++) {
+                if (ends_with(op->name, gpu_thread_name(dim))) {
                     break;
                 }
             }
@@ -203,7 +204,7 @@ class ReplaceForWithIf : public IRMutator {
 
             Stmt body = mutate(op->body);
 
-            Expr var = Variable::make(Int(32), "." + thread_names[dim]);
+            Expr var = Variable::make(Int(32), gpu_thread_name(dim));
             body = substitute(op->name, var + op->min, body);
 
             if (equal(op->extent, block_size.num_threads(dim))) {
@@ -322,7 +323,7 @@ private:
     }
 
     Stmt visit(const For *op) override {
-        bool is_thread_loop = CodeGen_GPU_Dev::is_gpu_thread_var(op->name);
+        bool is_thread_loop = op->for_type == ForType::GPUThread || op->for_type == ForType::GPULane;
         ScopedValue<bool> old_in_threads(in_threads, in_threads || is_thread_loop);
 
         // Set aside the allocations we've found so far.
@@ -627,7 +628,6 @@ private:
 
                 if (!may_merge_allocs_of_different_type &&
                     mem_allocs[free_spaces[i]].group[0].type != alloc.type) {
-                    // Types must also match for OpenGLCompute
                     continue;
                 }
 
@@ -649,7 +649,6 @@ private:
 
                 if (!may_merge_allocs_of_different_type &&
                     mem_allocs[free_spaces[i]].group[0].type != alloc.type) {
-                    // Types must also match for OpenGLCompute
                     continue;
                 }
 
@@ -760,7 +759,7 @@ public:
         // lifetimes, and then cluster the groups according to which
         // ones can share a single allocation. For cuda, opencl, and
         // similar we get one big combined allocation per memory
-        // type. For vulkan, openglcompute and direct3d, we also separate by
+        // type. For vulkan and direct3d, we also separate by
         // element type.
         map<pair<MemoryType, Type>, vector<AllocGroup>> clustered_allocs;
 
@@ -1034,8 +1033,7 @@ public:
         : device_api(d),
           thread_id_var_name(unique_name('t')),
           num_threads_var_name(unique_name('t')),
-          may_merge_allocs_of_different_type(device_api != DeviceAPI::OpenGLCompute &&
-                                             device_api != DeviceAPI::D3D12Compute &&
+          may_merge_allocs_of_different_type(device_api != DeviceAPI::D3D12Compute &&
                                              device_api != DeviceAPI::Vulkan &&
                                              device_api != DeviceAPI::WebGPU) {
     }
@@ -1143,21 +1141,21 @@ class ExtractRegisterAllocations : public IRMutator {
     }
 
     Expr visit(const Load *op) override {
-        string new_name = op->name;
-        if (alloc_renaming.contains(op->name)) {
-            new_name = alloc_renaming.get(op->name);
+        const string *new_name = alloc_renaming.find(op->name);
+        if (!new_name) {
+            new_name = &(op->name);
         }
-        return Load::make(op->type, new_name, mutate(op->index),
+        return Load::make(op->type, *new_name, mutate(op->index),
                           op->image, op->param, mutate(op->predicate),
                           op->alignment);
     }
 
     Stmt visit(const Store *op) override {
-        string new_name = op->name;
-        if (alloc_renaming.contains(op->name)) {
-            new_name = alloc_renaming.get(op->name);
+        const string *new_name = alloc_renaming.find(op->name);
+        if (!new_name) {
+            new_name = &(op->name);
         }
-        return Store::make(new_name, mutate(op->value), mutate(op->index),
+        return Store::make(*new_name, mutate(op->value), mutate(op->index),
                            op->param, mutate(op->predicate), op->alignment);
     }
 
@@ -1369,7 +1367,7 @@ class FuseGPUThreadLoopsSingleKernel : public IRMutator {
     ExtractSharedAndHeapAllocations &block_allocations;
 
     Stmt visit(const For *op) override {
-        if (ends_with(op->name, ".__block_id_x")) {
+        if (ends_with(op->name, gpu_block_name(0))) {
             Stmt body = op->body;
 
             // This is the innermost loop over blocks.
@@ -1410,17 +1408,17 @@ class FuseGPUThreadLoopsSingleKernel : public IRMutator {
             debug(3) << "Replaced for with if:\n"
                      << body << "\n\n";
 
-            // There is always a loop over thread_id_x
-            string thread_id = "." + thread_names[0];
+            // There is always a loop over the innermost thread dimension
+            string thread_id = gpu_thread_name(0);
             // Add back in any register-level allocations
             body = register_allocs.rewrap(body, thread_id);
             body = For::make(thread_id, 0, block_size_x, innermost_loop_type, op->partition_policy, op->device_api, body);
 
             // Rewrap the whole thing in other loops over threads
             for (int i = 1; i < block_size.threads_dimensions(); i++) {
-                thread_id = "." + thread_names[i];
+                thread_id = gpu_thread_name(i);
                 body = register_allocs.rewrap(body, thread_id);
-                body = For::make("." + thread_names[i], 0, block_size.num_threads(i),
+                body = For::make(thread_id, 0, block_size.num_threads(i),
                                  ForType::GPUThread, op->partition_policy, op->device_api, body);
             }
             thread_id.clear();
@@ -1455,14 +1453,15 @@ class FuseGPUThreadLoops : public IRMutator {
     using IRMutator::visit;
 
     Stmt visit(const For *op) override {
-        user_assert(!(CodeGen_GPU_Dev::is_gpu_thread_var(op->name)))
+        user_assert(!(op->for_type == ForType::GPUThread ||
+                      op->for_type == ForType::GPULane))
             << "Loops over GPU thread variable: \"" << op->name
             << "\" is outside of any loop over a GPU block variable. "
             << "This schedule is malformed. There must be a GPU block "
             << "variable, and it must reordered to be outside all GPU "
             << "thread variables.\n";
 
-        if (CodeGen_GPU_Dev::is_gpu_block_var(op->name)) {
+        if (op->for_type == ForType::GPUBlock) {
             // Do the analysis of thread block size and shared memory
             // usage.
             ExtractBlockSize block_size;
@@ -1501,7 +1500,7 @@ class ZeroGPULoopMins : public IRMutator {
                           (op->device_api == DeviceAPI::Vulkan);
 
         Stmt stmt = IRMutator::visit(op);
-        if (CodeGen_GPU_Dev::is_gpu_var(op->name) && !is_const_zero(op->min)) {
+        if (is_gpu(op->for_type) && !is_const_zero(op->min)) {
             op = stmt.as<For>();
             internal_assert(op);
             Expr adjusted = Variable::make(Int(32), op->name) + op->min;
@@ -1515,44 +1514,6 @@ public:
     ZeroGPULoopMins() = default;
 };
 
-class ValidateGPULoopNesting : public IRVisitor {
-    int gpu_block_depth = 0, gpu_thread_depth = 0;
-    string innermost_block_var, innermost_thread_var;
-
-    using IRVisitor::visit;
-
-    void visit(const For *op) override {
-        ScopedValue<string> old_innermost_block_var(innermost_block_var);
-        ScopedValue<string> old_innermost_thread_var(innermost_thread_var);
-        ScopedValue<int> old_gpu_block_depth(gpu_block_depth);
-        ScopedValue<int> old_gpu_thread_depth(gpu_thread_depth);
-
-        for (int i = 1; i <= 4; i++) {
-            if (ends_with(op->name, block_names[4 - i])) {
-                user_assert(i > gpu_block_depth)
-                    << "Invalid schedule: Loop over " << op->name
-                    << " cannot be inside of loop over " << innermost_block_var << "\n";
-                user_assert(gpu_thread_depth == 0)
-                    << "Invalid schedule: Loop over " << op->name
-                    << " cannot be inside of loop over " << innermost_thread_var << "\n";
-                innermost_block_var = op->name;
-                gpu_block_depth = i;
-            }
-            if (ends_with(op->name, thread_names[4 - i])) {
-                user_assert(i > gpu_thread_depth)
-                    << "Invalid schedule: Loop over " << op->name
-                    << " cannot be inside of loop over " << innermost_thread_var << "\n";
-                user_assert(gpu_block_depth > 0)
-                    << "Invalid schedule: Loop over " << op->name
-                    << " must be inside a loop over gpu blocks\n";
-                innermost_thread_var = op->name;
-                gpu_thread_depth = i;
-            }
-        }
-        IRVisitor::visit(op);
-    }
-};
-
 }  // namespace
 
 // Also used by InjectImageIntrinsics
@@ -1567,7 +1528,7 @@ class FindInnermostGPUBlock : public IRVisitor {
     using IRVisitor::visit;
 
     void visit(const For *op) override {
-        if (CodeGen_GPU_Dev::is_gpu_block_var(op->name)) {
+        if (op->for_type == ForType::GPUBlock) {
             // Set the last found GPU block to found_gpu_block.
             found_gpu_block = op;
         }
@@ -1608,7 +1569,7 @@ class NormalizeIfStatements : public IRMutator {
     bool inside_gpu_blocks = false;
 
     Stmt visit(const For *op) override {
-        if (!CodeGen_GPU_Dev::is_gpu_block_var(op->name)) {
+        if (op->for_type != ForType::GPUBlock) {
             return IRMutator::visit(op);
         }
         ScopedValue<bool> old_inside_gpu_blocks(inside_gpu_blocks, true);
@@ -1632,8 +1593,6 @@ class NormalizeIfStatements : public IRMutator {
 }  // namespace
 
 Stmt fuse_gpu_thread_loops(Stmt s) {
-    ValidateGPULoopNesting validate;
-    s.accept(&validate);
     // NormalizeIfStatements pushes the predicates between GPU blocks
     // into the innermost GPU block. FuseGPUThreadLoops would then
     // merge the predicate into the merged GPU thread.
