@@ -3,7 +3,7 @@
 #include <string>
 
 #include "CodeGen_Internal.h"
-#include "ExprUsesVar.h"
+#include "Function.h"
 #include "IRMutator.h"
 #include "IROperator.h"
 #include "InjectHostDevBufferCopies.h"
@@ -23,14 +23,38 @@ using std::vector;
 
 namespace {
 
-Stmt incr_active_threads(const Expr &profiler_state) {
+// All names that need to be unique, just in case someone does something
+// perverse like naming a func "profiler_instance".
+struct Names {
+    const std::string &pipeline_name;
+    std::string profiler_instance;
+    std::string profiler_local_sampling_token;
+    std::string profiler_shared_sampling_token;
+    std::string hvx_profiler_instance;
+    std::string profiler_func_names;
+    std::string profiler_func_stack_peak_buf;
+    std::string profiler_start_error_code;
+
+    Names(const std::string &pipeline_name)
+        : pipeline_name(pipeline_name),
+          profiler_instance(unique_name("profiler_instance")),
+          profiler_local_sampling_token(unique_name("profiler_local_sampling_token")),
+          profiler_shared_sampling_token(unique_name("profiler_shared_sampling_token")),
+          hvx_profiler_instance(unique_name("hvx_profiler_instance")),
+          profiler_func_names(unique_name("profiler_func_names")),
+          profiler_func_stack_peak_buf(unique_name("profiler_func_stack_peak_buf")),
+          profiler_start_error_code(unique_name("profiler_start_error_code")) {
+    }
+};
+
+Stmt incr_active_threads(const Expr &profiler_instance) {
     return Evaluate::make(Call::make(Int(32), "halide_profiler_incr_active_threads",
-                                     {profiler_state}, Call::Extern));
+                                     {profiler_instance}, Call::Extern));
 }
 
-Stmt decr_active_threads(const Expr &profiler_state) {
+Stmt decr_active_threads(const Expr &profiler_instance) {
     return Evaluate::make(Call::make(Int(32), "halide_profiler_decr_active_threads",
-                                     {profiler_state}, Call::Extern));
+                                     {profiler_instance}, Call::Extern));
 }
 
 Stmt acquire_sampling_token(const Expr &shared_token, const Expr &local_token) {
@@ -43,18 +67,6 @@ Stmt release_sampling_token(const Expr &shared_token, const Expr &local_token) {
                                      {shared_token, local_token}, Call::Extern));
 }
 
-Stmt activate_thread(const Stmt &s, const Expr &profiler_state) {
-    return Block::make({incr_active_threads(profiler_state),
-                        s,
-                        decr_active_threads(profiler_state)});
-}
-
-Stmt suspend_thread(const Stmt &s, const Expr &profiler_state) {
-    return Block::make({decr_active_threads(profiler_state),
-                        s,
-                        incr_active_threads(profiler_state)});
-}
-
 Stmt claim_sampling_token(const Stmt &s, const Expr &shared_token, const Expr &local_token) {
     return LetStmt::make(local_token.as<Variable>()->name,
                          Call::make(Handle(), Call::alloca, {Int(32).bytes()}, Call::Intrinsic),
@@ -70,37 +82,62 @@ public:
 
     vector<int> stack;  // What produce nodes are we currently inside of.
 
-    string pipeline_name;
+    const Names &names;
+    const map<string, Function> &env;
 
     bool in_fork = false;
     bool in_parallel = false;
     bool in_leaf_task = false;
 
-    InjectProfiling(const string &pipeline_name)
-        : pipeline_name(pipeline_name) {
+    InjectProfiling(const Names &names, const map<std::string, Function> &env)
+        : names(names), env(env) {
         stack.push_back(get_func_id("overhead"));
         // ID 0 is treated specially in the runtime as overhead
         internal_assert(stack.back() == 0);
 
+        waiting_on_tasks_id = get_func_id("waiting for parallel tasks to finish");
         malloc_id = get_func_id("halide_malloc");
         free_id = get_func_id("halide_free");
-        profiler_pipeline_state = Variable::make(Handle(), "profiler_pipeline_state");
-        profiler_state = Variable::make(Handle(), "profiler_state");
-        profiler_token = Variable::make(Int(32), "profiler_token");
-        profiler_local_sampling_token = Variable::make(Handle(), "profiler_local_sampling_token");
-        profiler_shared_sampling_token = Variable::make(Handle(), "profiler_shared_sampling_token");
+
+        profiler_instance = Variable::make(Handle(), names.profiler_instance);
+        profiler_local_sampling_token = Variable::make(Handle(), names.profiler_local_sampling_token);
+        profiler_shared_sampling_token = Variable::make(Handle(), names.profiler_shared_sampling_token);
     }
 
     map<int, uint64_t> func_stack_current;  // map from func id -> current stack allocation
     map<int, uint64_t> func_stack_peak;     // map from func id -> peak stack allocation
 
+    Stmt activate_thread(const Stmt &s) {
+        return activate_thread_helper(s, waiting_on_tasks_id);
+    }
+
+    Stmt activate_main_thread(const Stmt &s) {
+        // The same as a child task, except when we finish (but before the
+        // instances get popped), bill anything as overhead.
+        return activate_thread_helper(s, 0);
+    }
+
+    Stmt activate_thread_helper(const Stmt &s, int final_id) {
+        return Block::make({incr_active_threads(profiler_instance),
+                            unconditionally_set_current_func(stack.back()),
+                            s,
+                            decr_active_threads(profiler_instance),
+                            unconditionally_set_current_func(final_id)});
+    }
+
+    Stmt suspend_thread(const Stmt &s) {
+        return Block::make({decr_active_threads(profiler_instance),
+                            unconditionally_set_current_func(waiting_on_tasks_id),
+                            s,
+                            incr_active_threads(profiler_instance),
+                            unconditionally_set_current_func(stack.back())});
+    }
+
 private:
     using IRMutator::visit;
 
-    int malloc_id, free_id;
-    Expr profiler_pipeline_state;
-    Expr profiler_state;
-    Expr profiler_token;
+    int malloc_id, free_id, waiting_on_tasks_id;
+    Expr profiler_instance;
     Expr profiler_local_sampling_token;
     Expr profiler_shared_sampling_token;
 
@@ -119,10 +156,28 @@ private:
     bool profiling_memory = true;
 
     // Strip down the tuple name, e.g. f.0 into f
-    string normalize_name(const string &name) {
-        vector<string> v = split_string(name, ".");
-        internal_assert(!v.empty());
-        return v[0];
+    string normalize_name(const string &name) const {
+        size_t idx = name.find('.');
+        if (idx != std::string::npos) {
+            internal_assert(idx != 0);
+            return name.substr(0, idx);
+        } else {
+            return name;
+        }
+    }
+
+    Function lookup_function(const string &name) const {
+        auto it = env.find(name);
+        if (it != env.end()) {
+            return it->second;
+        }
+        string norm_name = normalize_name(name);
+        it = env.find(norm_name);
+        if (it != env.end()) {
+            return it->second;
+        }
+        internal_error << "No function in the environment found for name '" << name << "'.\n";
+        return {};
     }
 
     int get_func_id(const string &name) {
@@ -138,6 +193,12 @@ private:
         return idx;
     }
 
+    Stmt unconditionally_set_current_func(int id) {
+        Stmt s = Evaluate::make(Call::make(Int(32), "halide_profiler_set_current_func",
+                                           {profiler_instance, id, reinterpret(Handle(), cast<uint64_t>(0))}, Call::Extern));
+        return s;
+    }
+
     Stmt set_current_func(int id) {
         if (most_recently_set_func == id) {
             return Evaluate::make(0);
@@ -146,7 +207,7 @@ private:
         Expr last_arg = in_leaf_task ? profiler_local_sampling_token : reinterpret(Handle(), cast<uint64_t>(0));
         // This call gets inlined and becomes a single store instruction.
         Stmt s = Evaluate::make(Call::make(Int(32), "halide_profiler_set_current_func",
-                                           {profiler_state, profiler_token, id, last_arg}, Call::Extern));
+                                           {profiler_instance, id, last_arg}, Call::Extern));
 
         return s;
     }
@@ -184,8 +245,18 @@ private:
         return size;
     }
 
+    Expr visit(const Call *op) override {
+        if (op->is_intrinsic(Call::profiling_enable_instance_marker)) {
+            // We're out of the bounds query code. This instance should be
+            // tracked (including any samples taken before this point.
+            return Call::make(Int(32), "halide_profiler_enable_instance",
+                              {profiler_instance}, Call::Extern);
+        } else {
+            return IRMutator::visit(op);
+        }
+    }
+
     Stmt visit(const Allocate *op) override {
-        int idx = get_func_id(op->name);
 
         auto [new_extents, changed] = mutate_with_changes(op->extents);
         Expr condition = mutate(op->condition);
@@ -199,12 +270,19 @@ private:
         // always conditionally false. remove_dead_allocations() is called after
         // inject_profiling() so this is a possible scenario.
         if (!is_const_zero(size) && on_stack) {
+            int idx;
+            Function func = lookup_function(op->name);
+            if (func.should_not_profile()) {
+                idx = stack.back();  // Attribute the stack size contribution to the deepest _profiled_ func.
+            } else {
+                idx = get_func_id(op->name);
+            }
             const uint64_t *int_size = as_const_uint(size);
             internal_assert(int_size != nullptr);  // Stack size is always a const int
             func_stack_current[idx] += *int_size;
             func_stack_peak[idx] = std::max(func_stack_peak[idx], func_stack_current[idx]);
             debug(3) << "  Allocation on stack: " << op->name
-                     << "(" << size << ") in pipeline " << pipeline_name
+                     << "(" << size << ") in pipeline " << names.pipeline_name
                      << "; current: " << func_stack_current[idx]
                      << "; peak: " << func_stack_peak[idx] << "\n";
         }
@@ -212,13 +290,14 @@ private:
         vector<Stmt> tasks;
         bool track_heap_allocation = !is_const_zero(size) && !on_stack && profiling_memory;
         if (track_heap_allocation) {
+            int idx = get_func_id(op->name);
             debug(3) << "  Allocation on heap: " << op->name
                      << "(" << size << ") in pipeline "
-                     << pipeline_name << "\n";
+                     << names.pipeline_name << "\n";
 
             tasks.push_back(set_current_func(malloc_id));
             tasks.push_back(Evaluate::make(Call::make(Int(32), "halide_profiler_memory_allocate",
-                                                      {profiler_pipeline_state, idx, size}, Call::Extern)));
+                                                      {profiler_instance, idx, size}, Call::Extern)));
         }
 
         Stmt body = mutate(op->body);
@@ -245,8 +324,6 @@ private:
     }
 
     Stmt visit(const Free *op) override {
-        int idx = get_func_id(op->name);
-
         AllocSize alloc = func_alloc_sizes.get(op->name);
         internal_assert(alloc.size.type() == UInt(64));
         func_alloc_sizes.pop(op->name);
@@ -256,12 +333,13 @@ private:
         if (!is_const_zero(alloc.size)) {
             if (!alloc.on_stack) {
                 if (profiling_memory) {
-                    debug(3) << "  Free on heap: " << op->name << "(" << alloc.size << ") in pipeline " << pipeline_name << "\n";
+                    int idx = get_func_id(op->name);
+                    debug(3) << "  Free on heap: " << op->name << "(" << alloc.size << ") in pipeline " << names.pipeline_name << "\n";
 
                     vector<Stmt> tasks{
                         set_current_func(free_id),
                         Evaluate::make(Call::make(Int(32), "halide_profiler_memory_free",
-                                                  {profiler_pipeline_state, idx, alloc.size}, Call::Extern)),
+                                                  {profiler_instance, idx, alloc.size}, Call::Extern)),
                         stmt,
                         set_current_func(stack.back())};
 
@@ -271,9 +349,18 @@ private:
                 const uint64_t *int_size = as_const_uint(alloc.size);
                 internal_assert(int_size != nullptr);
 
+                int idx;
+                Function func = lookup_function(op->name);
+                if (func.should_not_profile()) {
+                    idx = stack.back();  // Attribute the stack size contribution to the deepest _profiled_ func.
+                } else {
+                    idx = get_func_id(op->name);
+                }
                 func_stack_current[idx] -= *int_size;
-                debug(3) << "  Free on stack: " << op->name << "(" << alloc.size << ") in pipeline " << pipeline_name
-                         << "; current: " << func_stack_current[idx] << "; peak: " << func_stack_peak[idx] << "\n";
+                debug(3) << "  Free on stack: " << op->name
+                         << "(" << alloc.size << ") in pipeline " << names.pipeline_name
+                         << "; current: " << func_stack_current[idx]
+                         << "; peak: " << func_stack_peak[idx] << "\n";
             }
         }
         return stmt;
@@ -283,11 +370,19 @@ private:
         int idx;
         Stmt body;
         if (op->is_producer) {
-            idx = get_func_id(op->name);
-            stack.push_back(idx);
-            Stmt set_current = set_current_func(idx);
-            body = Block::make(set_current, mutate(op->body));
-            stack.pop_back();
+            Function func = lookup_function(op->name);
+            if (func.should_not_profile()) {
+                body = mutate(op->body);
+                if (body.same_as(op->body)) {
+                    return op;
+                }
+            } else {
+                idx = get_func_id(op->name);
+                stack.push_back(idx);
+                Stmt set_current = set_current_func(idx);
+                body = Block::make(set_current, mutate(op->body));
+                stack.pop_back();
+            }
         } else {
             // At the beginning of the consume step, set the current task
             // back to the outer one.
@@ -305,7 +400,7 @@ private:
         } else if (const Acquire *a = s.as<Acquire>()) {
             s = Acquire::make(a->semaphore, a->count, visit_parallel_task(a->body));
         } else {
-            s = activate_thread(mutate(s), profiler_state);
+            s = activate_thread(mutate(s));
         }
         if (most_recently_set_func != old) {
             most_recently_set_func = -1;
@@ -315,13 +410,13 @@ private:
 
     Stmt visit(const Acquire *op) override {
         Stmt s = visit_parallel_task(op);
-        return suspend_thread(s, profiler_state);
+        return suspend_thread(s);
     }
 
     Stmt visit(const Fork *op) override {
         ScopedValue<bool> bind(in_fork, true);
         Stmt s = visit_parallel_task(op);
-        return suspend_thread(s, profiler_state);
+        return suspend_thread(s);
     }
 
     Stmt visit(const For *op) override {
@@ -338,7 +433,6 @@ private:
 
         bool leaf_task = false;
         if (update_active_threads) {
-            body = activate_thread(body, profiler_state);
 
             class ContainsParallelOrBlockingNode : public IRVisitor {
                 using IRVisitor::visit;
@@ -364,6 +458,8 @@ private:
             if (leaf_task) {
                 body = claim_sampling_token(body, profiler_shared_sampling_token, profiler_local_sampling_token);
             }
+
+            body = activate_thread(body);
         }
         ScopedValue<bool> bind_leaf_task(in_leaf_task, in_leaf_task || leaf_task);
 
@@ -383,9 +479,9 @@ private:
             // Get the profiler state pointer from scratch inside the
             // kernel. There will be a separate copy of the state on
             // the DSP that the host side will periodically query.
-            Expr get_state = Call::make(Handle(), "halide_profiler_get_state", {}, Call::Extern);
-            body = substitute("profiler_state", Variable::make(Handle(), "hvx_profiler_state"), body);
-            body = LetStmt::make("hvx_profiler_state", get_state, body);
+            Expr get_state = Call::make(Handle(), "halide_hexagon_remote_profiler_get_global_instance", {}, Call::Extern);
+            body = substitute(names.profiler_instance, Variable::make(Handle(), names.hvx_profiler_instance), body);
+            body = LetStmt::make(names.hvx_profiler_instance, get_state, body);
         } else if (op->device_api == DeviceAPI::None ||
                    op->device_api == DeviceAPI::Host) {
             body = mutate(body);
@@ -400,7 +496,7 @@ private:
         Stmt stmt = For::make(op->name, op->min, op->extent, op->for_type, op->partition_policy, op->device_api, body);
 
         if (update_active_threads) {
-            stmt = suspend_thread(stmt, profiler_state);
+            stmt = suspend_thread(stmt);
         }
 
         return stmt;
@@ -498,77 +594,84 @@ private:
 
 }  // namespace
 
-Stmt inject_profiling(Stmt s, const string &pipeline_name) {
-    InjectProfiling profiling(pipeline_name);
-    s = profiling.mutate(s);
+Stmt inject_profiling(const Stmt &stmt, const string &pipeline_name, const std::map<string, Function> &env) {
+    Names names(pipeline_name);
+
+    InjectProfiling profiling(names, env);
+    Stmt s = profiling.mutate(stmt);
 
     int num_funcs = (int)(profiling.indices.size());
 
-    Expr func_names_buf = Variable::make(Handle(), "profiling_func_names");
+    // TODO: unique_name all these strings
 
-    Expr start_profiler = Call::make(Int(32), "halide_profiler_pipeline_start",
-                                     {pipeline_name, num_funcs, func_names_buf}, Call::Extern);
+    Expr instance = Variable::make(Handle(), names.profiler_instance);
 
-    Expr get_state = Call::make(Handle(), "halide_profiler_get_state", {}, Call::Extern);
+    Expr func_names_buf = Variable::make(Handle(), names.profiler_func_names);
 
-    Expr get_pipeline_state = Call::make(Handle(), "halide_profiler_get_pipeline_state", {pipeline_name}, Call::Extern);
+    Expr start_profiler = Call::make(Int(32), "halide_profiler_instance_start",
+                                     {pipeline_name, num_funcs, func_names_buf, instance}, Call::Extern);
 
-    Expr profiler_token = Variable::make(Int(32), "profiler_token");
+    Expr profiler_start_error_code = Variable::make(Int(32), names.profiler_start_error_code);
 
     Expr stop_profiler = Call::make(Handle(), Call::register_destructor,
-                                    {Expr("halide_profiler_pipeline_end"), get_state}, Call::Intrinsic);
+                                    {Expr("halide_profiler_instance_end"), instance}, Call::Intrinsic);
 
     bool no_stack_alloc = profiling.func_stack_peak.empty();
     if (!no_stack_alloc) {
-        Expr func_stack_peak_buf = Variable::make(Handle(), "profiling_func_stack_peak_buf");
+        Expr func_stack_peak_buf = Variable::make(Handle(), names.profiler_func_stack_peak_buf);
 
-        Expr profiler_pipeline_state = Variable::make(Handle(), "profiler_pipeline_state");
         Stmt update_stack = Evaluate::make(Call::make(Int(32), "halide_profiler_stack_peak_update",
-                                                      {profiler_pipeline_state, func_stack_peak_buf}, Call::Extern));
+                                                      {instance, func_stack_peak_buf}, Call::Extern));
         s = Block::make(update_stack, s);
     }
 
-    Expr profiler_state = Variable::make(Handle(), "profiler_state");
-
-    s = activate_thread(s, profiler_state);
+    s = profiling.activate_main_thread(s);
 
     // Initialize the shared sampling token
-    Expr shared_sampling_token_var = Variable::make(Handle(), "profiler_shared_sampling_token");
+    Expr shared_sampling_token_var = Variable::make(Handle(), names.profiler_shared_sampling_token);
     Expr init_sampling_token =
         Call::make(Int(32), "halide_profiler_init_sampling_token", {shared_sampling_token_var, 0}, Call::Extern);
     s = Block::make({Evaluate::make(init_sampling_token), s});
-    s = LetStmt::make("profiler_shared_sampling_token",
+    s = LetStmt::make(names.profiler_shared_sampling_token,
                       Call::make(Handle(), Call::alloca, {Int(32).bytes()}, Call::Intrinsic), s);
 
-    s = LetStmt::make("profiler_pipeline_state", get_pipeline_state, s);
-    s = LetStmt::make("profiler_state", get_state, s);
     // If there was a problem starting the profiler, it will call an
     // appropriate halide error function and then return the
     // (negative) error code as the token.
-    s = Block::make(AssertStmt::make(profiler_token >= 0, profiler_token), s);
-    s = LetStmt::make("profiler_token", start_profiler, s);
+    s = Block::make(AssertStmt::make(profiler_start_error_code == 0, profiler_start_error_code), s);
+    s = LetStmt::make(names.profiler_start_error_code, start_profiler, s);
 
     if (!no_stack_alloc) {
         for (int i = num_funcs - 1; i >= 0; --i) {
-            s = Block::make(Store::make("profiling_func_stack_peak_buf",
+            s = Block::make(Store::make(names.profiler_func_stack_peak_buf,
                                         make_const(UInt(64), profiling.func_stack_peak[i]),
                                         i, Parameter(), const_true(), ModulusRemainder()),
                             s);
         }
-        s = Block::make(s, Free::make("profiling_func_stack_peak_buf"));
-        s = Allocate::make("profiling_func_stack_peak_buf", UInt(64),
+        s = Block::make(s, Free::make(names.profiler_func_stack_peak_buf));
+        s = Allocate::make(names.profiler_func_stack_peak_buf, UInt(64),
                            MemoryType::Auto, {num_funcs}, const_true(), s);
     }
 
     for (const auto &p : profiling.indices) {
-        s = Block::make(Store::make("profiling_func_names", p.first, p.second, Parameter(), const_true(), ModulusRemainder()), s);
+        s = Block::make(Store::make(names.profiler_func_names, p.first, p.second, Parameter(), const_true(), ModulusRemainder()), s);
     }
 
-    s = Block::make(s, Free::make("profiling_func_names"));
-    s = Allocate::make("profiling_func_names", Handle(),
+    s = Block::make(s, Free::make(names.profiler_func_names));
+    s = Allocate::make(names.profiler_func_names, Handle(),
                        MemoryType::Auto, {num_funcs}, const_true(), s);
     s = Block::make(Evaluate::make(stop_profiler), s);
 
+    // Allocate memory for the profiler instance state
+
+    // Check there isn't going to be end-of-struct padding to worry about.
+    static_assert((sizeof(halide_profiler_func_stats) & 7) == 0);
+
+    const int instance_size_bytes = sizeof(halide_profiler_instance_state) + num_funcs * sizeof(halide_profiler_func_stats);
+
+    s = Allocate::make(names.profiler_instance, UInt(64), MemoryType::Auto,
+                       {(instance_size_bytes + 7) / 8}, const_true(), s);
+
     // We have nested definitions of the sampling token
     s = uniquify_variable_names(s);
 
