@@ -624,6 +624,9 @@ public:
     bool &async();
     bool async() const;
 
+    Expr &ring_buffer();
+    Expr &ring_buffer() const;
+
     /** The list and order of dimensions used to store this
      * function. The first dimension in the vector corresponds to the
      * innermost dimension for storage (i.e. which dimension is
