@@ -7,7 +7,9 @@
  * exported in Halide.h. */
 
 #include "Bounds.h"
+#include "ConstantInterval.h"
 #include "IRMatch.h"
+#include "IRPrinter.h"
 #include "IRVisitor.h"
 #include "Scope.h"
 
@@ -28,17 +30,6 @@
 namespace Halide {
 namespace Internal {
 
-inline int64_t saturating_mul(int64_t a, int64_t b) {
-    int64_t result;
-    if (mul_with_overflow(64, a, b, &result)) {
-        return result;
-    } else if ((a > 0) == (b > 0)) {
-        return INT64_MAX;
-    } else {
-        return INT64_MIN;
-    }
-}
-
 class Simplify : public VariadicVisitor<Simplify, Expr, Stmt> {
     using Super = VariadicVisitor<Simplify, Expr, Stmt>;
 
@@ -47,80 +38,109 @@ public:
 
     struct ExprInfo {
         // We track constant integer bounds when they exist
-        // TODO: Use ConstantInterval?
-        int64_t min = 0, max = 0;
-        bool min_defined = false, max_defined = false;
+        ConstantInterval bounds;
         // And the alignment of integer variables
         ModulusRemainder alignment;
 
         void trim_bounds_using_alignment() {
             if (alignment.modulus == 0) {
-                min_defined = max_defined = true;
-                min = max = alignment.remainder;
+                bounds = ConstantInterval::single_point(alignment.remainder);
             } else if (alignment.modulus > 1) {
-                if (min_defined) {
+                if (bounds.min_defined) {
                     int64_t adjustment;
-                    bool no_overflow = sub_with_overflow(64, alignment.remainder, mod_imp(min, alignment.modulus), &adjustment);
+                    bool no_overflow = sub_with_overflow(64, alignment.remainder, mod_imp(bounds.min, alignment.modulus), &adjustment);
                     adjustment = mod_imp(adjustment, alignment.modulus);
                     int64_t new_min;
-                    no_overflow &= add_with_overflow(64, min, adjustment, &new_min);
+                    no_overflow &= add_with_overflow(64, bounds.min, adjustment, &new_min);
                     if (no_overflow) {
-                        min = new_min;
+                        bounds.min = new_min;
                     }
                 }
-                if (max_defined) {
+                if (bounds.max_defined) {
                     int64_t adjustment;
-                    bool no_overflow = sub_with_overflow(64, mod_imp(max, alignment.modulus), alignment.remainder, &adjustment);
+                    bool no_overflow = sub_with_overflow(64, mod_imp(bounds.max, alignment.modulus), alignment.remainder, &adjustment);
                     adjustment = mod_imp(adjustment, alignment.modulus);
                     int64_t new_max;
-                    no_overflow &= sub_with_overflow(64, max, adjustment, &new_max);
+                    no_overflow &= sub_with_overflow(64, bounds.max, adjustment, &new_max);
                     if (no_overflow) {
-                        max = new_max;
+                        bounds.max = new_max;
                     }
                 }
             }
 
-            if (min_defined && max_defined && min == max) {
+            if (bounds.is_single_point()) {
                 alignment.modulus = 0;
-                alignment.remainder = min;
+                alignment.remainder = bounds.min;
             }
+
+            if (bounds.is_bounded() && bounds.min > bounds.max) {
+                // Impossible, we must be in unreachable code. TODO: surface
+                // this to the simplify instance's in_unreachable flag.
+                bounds.max = bounds.min;
+            }
+        }
+
+        void cast_to(Type t) {
+            if ((!t.is_int() && !t.is_uint()) || (t.is_int() && t.bits() >= 32)) {
+                return;
+            }
+
+            // We've just done some infinite-integer operation on a bounded
+            // integer type, and we need to project the bounds and alignment
+            // back in-range.
+
+            if (!t.can_represent(bounds)) {
+                if (t.bits() >= 64) {
+                    // Just preserve any power-of-two factor in the modulus. When
+                    // alignment.modulus == 0, the value is some positive constant
+                    // representable as any 64-bit integer type, so there's no
+                    // wraparound.
+                    if (alignment.modulus > 0) {
+                        // This masks off all bits except for the lowest set one,
+                        // giving the largest power-of-two factor of a number.
+                        alignment.modulus &= -alignment.modulus;
+                        alignment.remainder = mod_imp(alignment.remainder, alignment.modulus);
+                    }
+                } else {
+                    // A narrowing integer cast that could possibly overflow adds
+                    // some unknown multiple of 2^bits
+                    alignment = alignment + ModulusRemainder(((int64_t)1 << t.bits()), 0);
+                }
+            }
+
+            // Truncate the bounds to the new type.
+            bounds.cast_to(t);
         }
 
         // Mix in existing knowledge about this Expr
         void intersect(const ExprInfo &other) {
-            if (min_defined && other.min_defined) {
-                min = std::max(min, other.min);
-            } else if (other.min_defined) {
-                min_defined = true;
-                min = other.min;
+            if (bounds < other.bounds || other.bounds < bounds) {
+                // Impossible. We must be in unreachable code. TODO: It might
+                // be nice to surface this to the simplify instance's
+                // in_unreachable flag, but we'd have to be sure that it's going
+                // to be caught at the right place.
+                return;
             }
-
-            if (max_defined && other.max_defined) {
-                max = std::min(max, other.max);
-            } else if (other.max_defined) {
-                max_defined = true;
-                max = other.max;
-            }
-
+            bounds = ConstantInterval::make_intersection(bounds, other.bounds);
             alignment = ModulusRemainder::intersect(alignment, other.alignment);
-
             trim_bounds_using_alignment();
         }
     };
 
     HALIDE_ALWAYS_INLINE
-    void clear_bounds_info(ExprInfo *b) {
+    void clear_expr_info(ExprInfo *b) {
         if (b) {
             *b = ExprInfo{};
         }
     }
 
 #if (LOG_EXPR_MUTATIONS || LOG_STMT_MUTATIONS)
-    static int debug_indent;
+    int debug_indent = 0;
 #endif
 
 #if LOG_EXPR_MUTATIONS
     Expr mutate(const Expr &e, ExprInfo *b) {
+        internal_assert(debug_indent >= 0);
         const std::string spaces(debug_indent, ' ');
         debug(1) << spaces << "Simplifying Expr: " << e << "\n";
         debug_indent++;
@@ -130,6 +150,19 @@ public:
             debug(1)
                 << spaces << "Before: " << e << "\n"
                 << spaces << "After:  " << new_e << "\n";
+            if (b) {
+                debug(1)
+                    << spaces << "Bounds: " << b->bounds << " " << b->alignment << "\n";
+                if (const int64_t *i = as_const_int(new_e)) {
+                    internal_assert(b->bounds.contains(*i)) << e << "\n"
+                                                            << new_e << "\n"
+                                                            << b->bounds;
+                } else if (const uint64_t *i = as_const_uint(new_e)) {
+                    internal_assert(b->bounds.contains(*i)) << e << "\n"
+                                                            << new_e << "\n"
+                                                            << b->bounds;
+                }
+            }
         }
         internal_assert(e.type() == new_e.type());
         return new_e;
@@ -258,7 +291,7 @@ public:
 
         std::vector<const Variable *> pop_list;
         std::vector<const Variable *> bounds_pop_list;
-        std::vector<Expr> truths, falsehoods;
+        std::set<Expr, IRDeepCompare> truths, falsehoods;
 
         void learn_false(const Expr &fact);
         void learn_true(const Expr &fact);
@@ -298,45 +331,45 @@ public:
     Stmt mutate_let_body(const Stmt &s, ExprInfo *) {
         return mutate(s);
     }
-    Expr mutate_let_body(const Expr &e, ExprInfo *bounds) {
-        return mutate(e, bounds);
+    Expr mutate_let_body(const Expr &e, ExprInfo *info) {
+        return mutate(e, info);
     }
 
     template<typename T, typename Body>
-    Body simplify_let(const T *op, ExprInfo *bounds);
+    Body simplify_let(const T *op, ExprInfo *info);
 
-    Expr visit(const IntImm *op, ExprInfo *bounds);
-    Expr visit(const UIntImm *op, ExprInfo *bounds);
-    Expr visit(const FloatImm *op, ExprInfo *bounds);
-    Expr visit(const StringImm *op, ExprInfo *bounds);
-    Expr visit(const Broadcast *op, ExprInfo *bounds);
-    Expr visit(const Cast *op, ExprInfo *bounds);
-    Expr visit(const Reinterpret *op, ExprInfo *bounds);
-    Expr visit(const Variable *op, ExprInfo *bounds);
-    Expr visit(const Add *op, ExprInfo *bounds);
-    Expr visit(const Sub *op, ExprInfo *bounds);
-    Expr visit(const Mul *op, ExprInfo *bounds);
-    Expr visit(const Div *op, ExprInfo *bounds);
-    Expr visit(const Mod *op, ExprInfo *bounds);
-    Expr visit(const Min *op, ExprInfo *bounds);
-    Expr visit(const Max *op, ExprInfo *bounds);
-    Expr visit(const EQ *op, ExprInfo *bounds);
-    Expr visit(const NE *op, ExprInfo *bounds);
-    Expr visit(const LT *op, ExprInfo *bounds);
-    Expr visit(const LE *op, ExprInfo *bounds);
-    Expr visit(const GT *op, ExprInfo *bounds);
-    Expr visit(const GE *op, ExprInfo *bounds);
-    Expr visit(const And *op, ExprInfo *bounds);
-    Expr visit(const Or *op, ExprInfo *bounds);
-    Expr visit(const Not *op, ExprInfo *bounds);
-    Expr visit(const Select *op, ExprInfo *bounds);
-    Expr visit(const Ramp *op, ExprInfo *bounds);
+    Expr visit(const IntImm *op, ExprInfo *info);
+    Expr visit(const UIntImm *op, ExprInfo *info);
+    Expr visit(const FloatImm *op, ExprInfo *info);
+    Expr visit(const StringImm *op, ExprInfo *info);
+    Expr visit(const Broadcast *op, ExprInfo *info);
+    Expr visit(const Cast *op, ExprInfo *info);
+    Expr visit(const Reinterpret *op, ExprInfo *info);
+    Expr visit(const Variable *op, ExprInfo *info);
+    Expr visit(const Add *op, ExprInfo *info);
+    Expr visit(const Sub *op, ExprInfo *info);
+    Expr visit(const Mul *op, ExprInfo *info);
+    Expr visit(const Div *op, ExprInfo *info);
+    Expr visit(const Mod *op, ExprInfo *info);
+    Expr visit(const Min *op, ExprInfo *info);
+    Expr visit(const Max *op, ExprInfo *info);
+    Expr visit(const EQ *op, ExprInfo *info);
+    Expr visit(const NE *op, ExprInfo *info);
+    Expr visit(const LT *op, ExprInfo *info);
+    Expr visit(const LE *op, ExprInfo *info);
+    Expr visit(const GT *op, ExprInfo *info);
+    Expr visit(const GE *op, ExprInfo *info);
+    Expr visit(const And *op, ExprInfo *info);
+    Expr visit(const Or *op, ExprInfo *info);
+    Expr visit(const Not *op, ExprInfo *info);
+    Expr visit(const Select *op, ExprInfo *info);
+    Expr visit(const Ramp *op, ExprInfo *info);
     Stmt visit(const IfThenElse *op);
-    Expr visit(const Load *op, ExprInfo *bounds);
-    Expr visit(const Call *op, ExprInfo *bounds);
-    Expr visit(const Shuffle *op, ExprInfo *bounds);
-    Expr visit(const VectorReduce *op, ExprInfo *bounds);
-    Expr visit(const Let *op, ExprInfo *bounds);
+    Expr visit(const Load *op, ExprInfo *info);
+    Expr visit(const Call *op, ExprInfo *info);
+    Expr visit(const Shuffle *op, ExprInfo *info);
+    Expr visit(const VectorReduce *op, ExprInfo *info);
+    Expr visit(const Let *op, ExprInfo *info);
     Stmt visit(const LetStmt *op);
     Stmt visit(const AssertStmt *op);
     Stmt visit(const For *op);
@@ -354,7 +387,7 @@ public:
     Stmt visit(const Atomic *op);
     Stmt visit(const HoistedStorage *op);
 
-    std::pair<std::vector<Expr>, bool> mutate_with_changes(const std::vector<Expr> &old_exprs, ExprInfo *bounds);
+    std::pair<std::vector<Expr>, bool> mutate_with_changes(const std::vector<Expr> &old_exprs);
 };
 
 }  // namespace Internal
