@@ -21,9 +21,35 @@
 #endif
 
 #ifdef _MSC_VER
+#define NOMINMAX
+#define WIN32_LEAN_AND_MEAN
 #include <intrin.h>
+#include <windows.h>
 #endif  // _MSC_VER
 
+#ifdef __APPLE__
+#include <mach/machine.h>
+#include <sys/sysctl.h>
+#include <sys/types.h>
+#endif
+
+#if defined(__linux__) && (defined(__arm__) || defined(__aarch64__))
+#include <asm/hwcap.h>
+#include <sys/auxv.h>
+#ifndef HWCAP_ASIMDHP
+#define HWCAP_ASIMDHP 0
+#endif
+#ifndef HWCAP_ASIMDDP
+#define HWCAP_ASIMDDP 0
+#endif
+#ifndef HWCAP_SVE
+#define HWCAP_SVE 0
+#endif
+#ifndef HWCAP2_SVE2
+#define HWCAP2_SVE2 0
+#endif
+#endif
+
 namespace Halide {
 
 using std::string;
@@ -31,13 +57,14 @@ using std::vector;
 
 namespace {
 
-#ifdef _MSC_VER
-static void cpuid(int info[4], int infoType, int extra) {
+#if defined(_M_IX86) || defined(_M_AMD64)
+
+void cpuid(int info[4], int infoType, int extra) {
     __cpuidex(info, infoType, extra);
 }
-#else
 
-#if defined(__x86_64__) || defined(__i386__)
+#elif defined(__x86_64__) || defined(__i386__)
+
 // CPU feature detection code taken from ispc
 // (https://github.com/ispc/ispc/blob/master/builtins/dispatch.ll)
 
@@ -47,10 +74,10 @@ void cpuid(int info[4], int infoType, int extra) {
         : "=a"(info[0]), "=b"(info[1]), "=c"(info[2]), "=d"(info[3])
         : "0"(infoType), "2"(extra));
 }
-#endif
+
 #endif
 
-#if defined(__x86_64__) || defined(__i386__) || defined(_MSC_VER)
+#if defined(__x86_64__) || defined(__i386__) || defined(_M_IX86) || defined(_M_AMD64)
 
 enum class VendorSignatures {
     Unknown,
@@ -143,6 +170,29 @@ Target::Processor get_amd_processor(unsigned family, unsigned model, bool have_s
 
 #endif  // defined(__x86_64__) || defined(__i386__) || defined(_MSC_VER)
 
+#ifdef __APPLE__
+
+template<typename T>
+std::optional<T> getsysctl(const char *name) {
+    T value;
+    size_t size = sizeof(value);
+    if (sysctlbyname(name, &value, &size, nullptr, 0)) {
+        return std::nullopt;
+    }
+    return std::make_optional(value);
+}
+
+bool sysctl_is_set(const char *name) {
+    return getsysctl<int>(name).value_or(0);
+}
+
+bool is_armv7s() {
+    return getsysctl<cpu_type_t>("hw.cputype") == CPU_TYPE_ARM &&
+           getsysctl<cpu_subtype_t>("hw.cpusubtype") == CPU_SUBTYPE_ARM_V7S;
+}
+
+#endif  // __APPLE__
+
 Target calculate_host_target() {
     Target::OS os = Target::OSUnknown;
 #ifdef __linux__
@@ -164,8 +214,66 @@ Target calculate_host_target() {
 #if __riscv
     Target::Arch arch = Target::RISCV;
 #else
-#if defined(__arm__) || defined(__aarch64__)
+#if defined(__arm__) || defined(__aarch64__) || defined(_M_ARM64) || defined(_M_ARM64EC)
     Target::Arch arch = Target::ARM;
+
+#ifdef __APPLE__
+    if (is_armv7s()) {
+        initial_features.push_back(Target::ARMv7s);
+    }
+
+    if (sysctl_is_set("hw.optional.arm.FEAT_DotProd")) {
+        initial_features.push_back(Target::ARMDotProd);
+    }
+
+    if (sysctl_is_set("hw.optional.arm.FEAT_FP16")) {
+        initial_features.push_back(Target::ARMFp16);
+    }
+#endif
+
+#ifdef __linux__
+    unsigned long hwcaps = getauxval(AT_HWCAP);
+    unsigned long hwcaps2 = getauxval(AT_HWCAP2);
+
+    if (hwcaps & HWCAP_ASIMDDP) {
+        initial_features.push_back(Target::ARMDotProd);
+    }
+
+    if (hwcaps & HWCAP_ASIMDHP) {
+        initial_features.push_back(Target::ARMFp16);
+    }
+
+    if (hwcaps & HWCAP_SVE) {
+        initial_features.push_back(Target::SVE);
+    }
+
+    if (hwcaps2 & HWCAP2_SVE2) {
+        initial_features.push_back(Target::SVE2);
+    }
+#endif
+
+#ifdef _MSC_VER
+
+    // Magic value from: https://github.com/dotnet/runtime/blob/7e977dcbe5efaeec2c75ed0c3e200c85b2e55522/src/native/minipal/cpufeatures.c#L19
+#define PF_ARM_SVE_INSTRUCTIONS_AVAILABLE (46)
+
+    // This is the strategy used by Google's cpuinfo library for
+    // detecting fp16 arithmetic support on Windows.
+    if (!IsProcessorFeaturePresent(PF_FLOATING_POINT_EMULATED) &&
+        IsProcessorFeaturePresent(PF_ARM_FMAC_INSTRUCTIONS_AVAILABLE)) {
+        initial_features.push_back(Target::ARMFp16);
+    }
+
+    if (IsProcessorFeaturePresent(PF_ARM_V82_DP_INSTRUCTIONS_AVAILABLE)) {
+        initial_features.push_back(Target::ARMDotProd);
+    }
+
+    if (IsProcessorFeaturePresent(PF_ARM_SVE_INSTRUCTIONS_AVAILABLE)) {
+        initial_features.push_back(Target::SVE);
+    }
+
+#endif
+
 #else
 #if defined(__powerpc__) && (defined(__FreeBSD__) || defined(__linux__))
     Target::Arch arch = Target::POWERPC;
@@ -251,6 +359,8 @@ Target calculate_host_target() {
         // Call cpuid with eax=7, ecx=0
         int info2[4];
         cpuid(info2, 7, 0);
+        int info3[4];
+        cpuid(info3, 7, 1);
         const uint32_t avx2 = 1U << 5;
         const uint32_t avx512f = 1U << 16;
         const uint32_t avx512dq = 1U << 17;
@@ -283,8 +393,6 @@ Target calculate_host_target() {
 
                 const uint32_t avxvnni = 1U << 4;     // avxvnni (note, not avx512vnni) result in eax
                 const uint32_t avx512bf16 = 1U << 5;  // bf16 result in eax, with cpuid(eax=7, ecx=1)
-                int info3[4];
-                cpuid(info3, 7, 1);
                 // TODO: port to family/model -based detection.
                 if ((info3[0] & avxvnni) == avxvnni &&
                     (info3[0] & avx512bf16) == avx512bf16) {
@@ -292,7 +400,40 @@ Target calculate_host_target() {
                 }
             }
         }
+
+        // AVX10 converged vector instructions.
+        const uint32_t avx10 = 1U << 19;
+        if (info2[3] & avx10) {
+            int info_avx10[4];
+            cpuid(info_avx10, 0x24, 0x0);
+
+            // This checks that the AVX10 version is greater than zero.
+            // It isn't really needed as for now only one version exists, but
+            // the docs indicate bits 0:7 of EBX should be >= 0 so...
+            if ((info[1] & 0xff) >= 1) {
+                initial_features.push_back(Target::AVX10_1);
+
+                const uint32_t avx10_128 = 1U << 16;
+                const uint32_t avx10_256 = 1U << 17;
+                const uint32_t avx10_512 = 1U << 18;
+                // Choose the maximum one that is available.
+                if (info[1] & avx10_512) {
+                    vector_bits = 512;
+                } else if (info[1] & avx10_256) {
+                    vector_bits = 256;
+                } else if (info[1] & avx10_128) {  // Not clear it is worth turning on AVX10 for this case.
+                    vector_bits = 128;
+                }
+            }
+        }
+
+        // APX register extensions, etc.
+        const uint32_t apx = 1U << 21;
+        if (info3[3] & apx) {
+            initial_features.push_back(Target::X86APX);
+        }
     }
+
 #endif
 #endif
 #endif
@@ -305,6 +446,7 @@ bool is_using_hexagon(const Target &t) {
             t.has_feature(Target::HVX_v62) ||
             t.has_feature(Target::HVX_v65) ||
             t.has_feature(Target::HVX_v66) ||
+            t.has_feature(Target::HVX_v68) ||
             t.has_feature(Target::HexagonDma) ||
             t.arch == Target::Hexagon);
 }
@@ -322,6 +464,9 @@ int get_hvx_lower_bound(const Target &t) {
     if (t.has_feature(Target::HVX_v66)) {
         return 66;
     }
+    if (t.has_feature(Target::HVX_v68)) {
+        return 68;
+    }
     return 60;
 }
 
@@ -498,7 +643,6 @@ const std::map<std::string, Target::Feature> feature_name_map = {
     {"cl_doubles", Target::CLDoubles},
     {"cl_half", Target::CLHalf},
     {"cl_atomics64", Target::CLAtomics64},
-    {"openglcompute", Target::OpenGLCompute},
     {"egl", Target::EGL},
     {"user_context", Target::UserContext},
     {"profile", Target::Profile},
@@ -511,6 +655,7 @@ const std::map<std::string, Target::Feature> feature_name_map = {
     {"hvx_v62", Target::HVX_v62},
     {"hvx_v65", Target::HVX_v65},
     {"hvx_v66", Target::HVX_v66},
+    {"hvx_v68", Target::HVX_v68},
     {"fuzz_float_stores", Target::FuzzFloatStores},
     {"soft_float_abi", Target::SoftFloatABI},
     {"msan", Target::MSAN},
@@ -557,6 +702,8 @@ const std::map<std::string, Target::Feature> feature_name_map = {
     {"vk_v12", Target::VulkanV12},
     {"vk_v13", Target::VulkanV13},
     {"semihosting", Target::Semihosting},
+    {"avx10_1", Target::AVX10_1},
+    {"x86apx", Target::X86APX},
     // NOTE: When adding features to this map, be sure to update PyEnums.cpp as well.
 };
 
@@ -858,6 +1005,7 @@ void Target::validate_features() const {
                                 HVX_v62,
                                 HVX_v65,
                                 HVX_v66,
+                                HVX_v68,
                                 NoNEON,
                                 POWER_ARCH_2_07,
                                 RVV,
@@ -983,9 +1131,6 @@ bool Target::supported() const {
 #if !defined(WITH_METAL)
     bad |= has_feature(Target::Metal);
 #endif
-#if !defined(WITH_OPENGLCOMPUTE)
-    bad |= has_feature(Target::OpenGLCompute);
-#endif
 #if !defined(WITH_D3D12)
     bad |= has_feature(Target::D3D12Compute);
 #endif
@@ -1059,7 +1204,6 @@ bool Target::has_gpu_feature() const {
             has_feature(OpenCL) ||
             has_feature(Metal) ||
             has_feature(D3D12Compute) ||
-            has_feature(OpenGLCompute) ||
             has_feature(Vulkan) ||
             has_feature(WebGPU));
 }
@@ -1118,14 +1262,12 @@ bool Target::supports_type(const Type &t) const {
     if (t.bits() == 64) {
         if (t.is_float()) {
             return (!has_feature(Metal) &&
-                    !has_feature(OpenGLCompute) &&
                     !has_feature(D3D12Compute) &&
                     (!has_feature(Target::OpenCL) || has_feature(Target::CLDoubles)) &&
                     (!has_feature(Vulkan) || has_feature(Target::VulkanFloat64)) &&
                     !has_feature(WebGPU));
         } else {
             return (!has_feature(Metal) &&
-                    !has_feature(OpenGLCompute) &&
                     !has_feature(D3D12Compute) &&
                     (!has_feature(Vulkan) || has_feature(Target::VulkanInt64)) &&
                     !has_feature(WebGPU));
@@ -1157,8 +1299,6 @@ bool Target::supports_type(const Type &t, DeviceAPI device) const {
         // Shader Model 5.x can optionally support double-precision; 64-bit int
         // types are not supported.
         return t.bits() < 64;
-    } else if (device == DeviceAPI::OpenGLCompute) {
-        return t.bits() < 64;
     } else if (device == DeviceAPI::Vulkan) {
         if (t.is_float() && t.bits() == 64) {
             return has_feature(Target::VulkanFloat64);
@@ -1214,9 +1354,6 @@ DeviceAPI Target::get_required_device_api() const {
     if (has_feature(Target::OpenCL)) {
         return DeviceAPI::OpenCL;
     }
-    if (has_feature(Target::OpenGLCompute)) {
-        return DeviceAPI::OpenGLCompute;
-    }
     if (has_feature(Target::Vulkan)) {
         return DeviceAPI::Vulkan;
     }
@@ -1232,8 +1369,6 @@ Target::Feature target_feature_for_device_api(DeviceAPI api) {
         return Target::CUDA;
     case DeviceAPI::OpenCL:
         return Target::OpenCL;
-    case DeviceAPI::OpenGLCompute:
-        return Target::OpenGLCompute;
     case DeviceAPI::Metal:
         return Target::Metal;
     case DeviceAPI::Hexagon:
@@ -1333,7 +1468,6 @@ bool Target::get_runtime_compatible_target(const Target &other, Target &result)
         Metal,
         NoNEON,
         OpenCL,
-        OpenGLCompute,
         Vulkan,
         WebGPU,
 
@@ -1351,6 +1485,7 @@ bool Target::get_runtime_compatible_target(const Target &other, Target &result)
         HVX_v62,
         HVX_v65,
         HVX_v66,
+        HVX_v68,
         VulkanV10,
         VulkanV12,
         VulkanV13,
@@ -1495,6 +1630,9 @@ bool Target::get_runtime_compatible_target(const Target &other, Target &result)
     if (hvx_version < 66) {
         output.features.reset(HVX_v66);
     }
+    if (hvx_version < 68) {
+        output.features.reset(HVX_v68);
+    }
 
     result = output;
     return true;
@@ -1550,7 +1688,7 @@ void target_test() {
     internal_assert(with_vector_bits.vector_bits == 512) << "Vector bits not populated in constructor.\n";
     internal_assert(Target(with_vector_bits.to_string()).vector_bits == 512) << "Vector bits not round tripped properly.\n";
 
-    std::cout << "Target test passed" << std::endl;
+    std::cout << "Target test passed\n";
 }
 
 }  // namespace Internal
