@@ -651,12 +651,12 @@ void generate_schedule(const std::vector<Function> &outputs,
         }
     }
 
-    aslog(1) << "Number of states added: " << stats.num_states_added << '\n';
-    aslog(1) << "Number of featurizations computed: " << stats.num_featurizations << '\n';
-    aslog(1) << "Number of memoization hits: " << stats.num_memoization_hits << '\n';
-    aslog(1) << "Number of memoization misses: " << stats.num_memoization_misses << '\n';
-    aslog(1) << "Number of block memoization hits: " << stats.num_block_memoization_hits << '\n';
-    aslog(1) << "Number of block memoization misses: " << stats.num_block_memoization_misses << '\n';
+    aslog(1) << "Number of states added: " << stats.num_states_added << "\n";
+    aslog(1) << "Number of featurizations computed: " << stats.num_featurizations << "\n";
+    aslog(1) << "Number of memoization hits: " << stats.num_memoization_hits << "\n";
+    aslog(1) << "Number of memoization misses: " << stats.num_memoization_misses << "\n";
+    aslog(1) << "Number of block memoization hits: " << stats.num_block_memoization_hits << "\n";
+    aslog(1) << "Number of block memoization misses: " << stats.num_block_memoization_misses << "\n";
     aslog(1) << "Total featurization time (ms): " << stats.total_featurization_time() << "\n";
     aslog(1) << "Average featurization time (ms): " << stats.average_featurization_time() << "\n";
     aslog(1) << "Total enqueue time (ms): " << stats.total_enqueue_time() << "\n";
@@ -667,14 +667,14 @@ void generate_schedule(const std::vector<Function> &outputs,
     aslog(1) << "Total filter thread tiles time (ms): " << stats.total_filter_thread_tiles_time() << "\n";
     aslog(1) << "Total filter parallel tiles time (ms): " << stats.total_filter_parallel_tiles_time() << "\n";
 
-    aslog(1) << "Number of schedules evaluated by cost model: " << stats.num_schedules_enqueued << '\n';
-    aslog(1) << "Number of tilings generated: " << stats.num_tilings_generated << '\n';
-    aslog(1) << "Number of tilings accepted: " << stats.num_tilings_accepted << '\n';
+    aslog(1) << "Number of schedules evaluated by cost model: " << stats.num_schedules_enqueued << "\n";
+    aslog(1) << "Number of tilings generated: " << stats.num_tilings_generated << "\n";
+    aslog(1) << "Number of tilings accepted: " << stats.num_tilings_accepted << "\n";
     aslog(1) << "Total cost model evaluation time (ms): " << stats.total_cost_model_evaluation_time() << "\n";
     aslog(1) << "Average cost model evaluation time (ms): " << stats.average_cost_model_evaluation_time() << "\n";
     std::chrono::duration<double> total_time = timer.elapsed();
     aslog(1) << "Time taken for autoscheduler (s): "
-             << std::chrono::duration_cast<std::chrono::milliseconds>(total_time).count() / 1000.0 << '\n';
+             << std::chrono::duration_cast<std::chrono::milliseconds>(total_time).count() / 1000.0 << "\n";
 }
 
 struct Anderson2021 {
