@@ -7,7 +7,6 @@
 #include "HalideRuntimeHexagonHost.h"
 #include "HalideRuntimeMetal.h"
 #include "HalideRuntimeOpenCL.h"
-#include "HalideRuntimeOpenGLCompute.h"
 #include "HalideRuntimeQurt.h"
 #include "HalideRuntimeVulkan.h"
 #include "HalideRuntimeWebGPU.h"
@@ -86,9 +85,11 @@ extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_error_param_too_small_u64,
     (void *)&halide_error_requirement_failed,
     (void *)&halide_error_specialize_fail,
+    (void *)&halide_error_split_factor_not_positive,
     (void *)&halide_error_unaligned_host_ptr,
     (void *)&halide_error_storage_bound_too_small,
     (void *)&halide_error_device_crop_failed,
+    (void *)&halide_error_vscale_invalid,
     (void *)&halide_float16_bits_to_double,
     (void *)&halide_float16_bits_to_float,
     (void *)&halide_free,
@@ -160,19 +161,14 @@ extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_opencl_set_device_type,
     (void *)&halide_opencl_set_platform_name,
     (void *)&halide_opencl_wrap_cl_mem,
-    (void *)&halide_opengl_create_context,
-    (void *)&halide_opengl_get_proc_address,
-    (void *)&halide_openglcompute_device_interface,
-    (void *)&halide_openglcompute_initialize_kernels,
-    (void *)&halide_openglcompute_finalize_kernels,
-    (void *)&halide_openglcompute_run,
     (void *)&halide_pointer_to_string,
     (void *)&halide_print,
     (void *)&halide_profiler_get_pipeline_state,
     (void *)&halide_profiler_get_state,
+    (void *)&halide_profiler_instance_start,
+    (void *)&halide_profiler_instance_end,
     (void *)&halide_profiler_memory_allocate,
     (void *)&halide_profiler_memory_free,
-    (void *)&halide_profiler_pipeline_start,
     (void *)&halide_profiler_report,
     (void *)&halide_profiler_reset,
     (void *)&halide_profiler_stack_peak_update,
@@ -200,7 +196,7 @@ extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_set_trace_file,
     (void *)&halide_shutdown_thread_pool,
     (void *)&halide_shutdown_trace,
-    (void *)&halide_sleep_ms,
+    (void *)&halide_sleep_us,
     (void *)&halide_spawn_thread,
     (void *)&halide_start_clock,
     (void *)&halide_start_timer_chain,
