@@ -18,6 +18,13 @@ int main(int argc, char *argv[]) {
 
     Target target = get_jit_target_from_environment();
 
+    // Disable the Vulkan validation layer or we'll leak
+    // https://github.com/halide/Halide/issues/8290
+    if (target.has_feature(Target::Vulkan)) {
+        char clear_env_var[] = "VK_INSTANCE_LAYERS=";
+        putenv(clear_env_var);
+    }
+
     // We need to hook the default handler too, to catch the frees done by release_all
     JITHandlers handlers;
     handlers.custom_print = halide_print;
@@ -58,7 +65,8 @@ int main(int argc, char *argv[]) {
 
     int ret = tracker.validate_gpu_object_lifetime(true /* allow_globals */, true /* allow_none */, 1 /* max_globals */);
     if (ret != 0) {
-        return ret;
+        fprintf(stderr, "validate_gpu_object_lifetime() failed\n");
+        return 1;
     }
 
     printf("Success!\n");
